""" 
This module holds utilties that can be reused across other modules
"""

# Standard library imports
from datetime import datetime
from itertools import cycle, repeat, tee
from pathlib import Path
from typing import Iterable, Tuple
import glob
import argparse

# Third-party library imports
import h5py
import numpy as np
import yaml
from PIL import Image
from scipy import ndimage
import skimage
import skimage.io as skio

# Local imports
from recon3d.types import *

# import recon3d.feature_analysis as fa
# from recon3d.feature_analysis import SemanticImageStack


# def instance_to_ndarray(data: InstanceImageStack) -> np.ndarray:
#     """Extract the array data within the Instance ImageStack object"""

# return data.data # Trivial, do we need a function?


# def hdf_dataset_to_npy(hdf_path:Path, hdf_dataset_location: str, save_path: Path):
#     # read
#     with h5py.File(hdf_path, "r") as f:
#         data = np.squeeze(f[hdf_dataset_location][:])

#     instance_to_ndarray
#     np.save(save_path, data)


def binary_with_pores_to_semantic(input_path: Path, output_path: Path) -> dict:
    """
    Convert a folder of segmented/binarized TIFF images to a Semantic Image Stack.

    This function is designed for a specific use case (NOMAD, pores in AM tensile bars)
    and converts binary images with metal as 1 and all else as 0 into a semantic image stack.

    Parameters
    ----------
    input_path : Path
        Folder containing binary images with metal as 1 and all else as 0.
        Figure 1:
        *---*---*---*---*
        | 1 | 1 | 1 | 1 |
        *---*---*---*---*
        | 1 | 0 | 1 | 0 |
        *---*---*---*---*
        | 1 | 1 | 1 | 0 |
        *---*---*---*---*
    output_path : Path
        Directory where the semantic images will be saved.
        Figure 2:
        *---*---*---*---*
        | 1 | 1 | 1 | 1 |
        *---*---*---*---*
        | 1 | 2 | 1 | 0 |
        *---*---*---*---*
        | 1 | 1 | 1 | 0 |
        *---*---*---*---*

    Returns
    -------
    dict
        Dictionary of class labels:
            class_labels:
                air:
                    value: 0
                metal:
                    value: 1
                pore:
                    value: 2

    Raises
    ------
    ValueError
        If the input images contain more than two phases.

    Examples
    --------
    >>> input_path = Path("path/to/binary_images")
    >>> output_path = Path("path/to/save/semantic_images")
    >>> binary_with_pores_to_semantic(input_path, output_path)
    {'class_labels': {'air': {'value': 0}, 'metal': {'value': 1}, 'pore': {'value': 2}}}

    """

    # read in binary image stack
    bw_data = read_images(input_path)

    # check data is binary, only 0 or 1
    phase_ids = np.unique(bw_data)
    n_phases = len(phase_ids)

    if n_phases != 2:
        raise ValueError(
            f"Only two phases expected in segmented image, {n_phases} phase found"
        )

    # elif n_phases == 2:
    # TODO remove hardcodes and impart flexibility from here to ut.ndarray_to_img call
    else:  # images are binary

        # preallocate the output data
        output_data = np.zeros(shape=bw_data.shape, dtype=np.int8)

        # fill holes and assign as 'metal', 1
        print("\tIsolating Sample...")
        sample = ndimage.binary_fill_holes(bw_data.astype(np.bool_))
        np.place(output_data, sample, 1)

        # isolate holes within 'metal', assign as 'pore', 2
        print("\tIsolating Voids...")
        voids = np.logical_xor(sample, bw_data)
        np.place(output_data, voids, 2)

        # thus, everything else is 'air', 0

    class_labels = {
        "class_labels": {
            "air": {"value": 0},
            "metal": {"value": 1},
            "pore": {"value": 2},
        }
    }

    ndarray_to_img(
        data=output_data,
        slice_axis=CartesianAxis3D.Z,
        parent_dir=output_path,
        folder_name="",
    )

    return class_labels


def validate_yml(yml_input_file: Path, cli_entry_point: str) -> tuple[Path, Path]:
    """
    Verify that the YAML file contains the required arguments.

    This function checks if the specified YAML file contains the necessary arguments
    and validates the CLI entry point. It also verifies the existence of input and output directories.

    Parameters
    ----------
    yml_input_file : Path
        The path to the YAML input file.
    cli_entry_point : str
        The CLI entry point to validate against the YAML file.

    Returns
    -------
    tuple[Path, Path, dict]
        A tuple containing:
        - Path to the input directory.
        - Path to the output directory.
        - Dictionary of YAML values.

    Raises
    ------
    ValueError
        If the CLI entry point is not found in the YAML file or if the input directory does not exist.

    Examples
    --------
    >>> yml_input_file = Path("path/to/input.yml")
    >>> cli_entry_point = "process_images"
    >>> validate_yml(yml_input_file, cli_entry_point)
    (PosixPath('path/to/images'), PosixPath('path/to/output'), {'cli_entry_points': ['process_images'], 'image_dir': 'path/to/images', 'out_dir': 'path/to/output'})
    """

    print(f"Processing specification file: {yml_input_file}")
    yml_vals = yaml_to_dict(yml_input_file)

    # check cli_entry_points is valid
    if cli_entry_point not in yml_vals["cli_entry_points"]:
        raise ValueError(
            f"""Error. Incorrect yml format. 
                         This function requires the "cli_entry_points" key to contain {cli_entry_point}, 
                         but currently contains the following options: {yml_vals["cli_entry_points"]} """
        )

    path_input = Path(yml_vals["image_dir"]).expanduser()
    if not path_input.is_dir():
        raise ValueError(f"Error, 'image_dir', {path_input} not found.")
    print(f"Input path: {path_input}")

    path_output = Path(yml_vals["out_dir"]).expanduser()
    path_output.mkdir(parents=True, exist_ok=True)
    print(f"Output path: {path_output}")

    return (path_input, path_output, yml_vals)


def binarize(data: np.ndarray, val: int) -> np.ndarray:
    """
    Binarize the data based on a specified value.

    This function converts the values within the data matching the specified value to 1,
    and all other values to 0.

    Parameters
    ----------
    data : np.ndarray
        The input array to be binarized.
    val : int
        The value in the data to be binarized (converted to 1).

    Returns
    -------
    np.ndarray
        The binarized array with the same shape as the input data.

    Examples
    --------
    >>> data = np.array([[1, 2, 3], [4, 1, 6], [7, 8, 1]])
    >>> binarize(data, 1)
    array([[1, 0, 0],
           [0, 1, 0],
           [0, 0, 1]], dtype=int8)
    """

    bw_data = np.zeros(shape=data.shape, dtype=np.int8)
    np.place(bw_data, data == val, 1)

    return bw_data


def semantic_to_binary(yml_input_file: Path) -> bool:
    """
    Convert a semantic image stack to a binary image stack using a YAML configuration file.

    This function reads a YAML file to prepare a binary image stack from a semantic image stack.
    The selected class from the semantic stack is binarized based on the specified value.

    Parameters
    ----------
    yml_input_file : Path
        The path to the YAML input file containing configuration settings.

    Returns
    -------
    bool
        True if the binary image stack was successfully created, False otherwise.

    Examples
    --------
    >>> yml_input_file = Path("path/to/config.yml")
    >>> semantic_to_binary(yml_input_file)
    True
    """

    [input_path, output_path, params] = validate_yml(
        yml_input_file, "semantic_to_binary"
    )

    # params = ut.yaml_to_dict(yml_input_file)

    sel_class = params["selected_class"]
    sel_class_value = params["class_labels"][sel_class]["value"]

    input_data = read_images(input_path)

    bw_data = binarize(input_data, sel_class_value)

    ndarray_to_img(
        data=bw_data,
        slice_axis=CartesianAxis3D.Z,
        parent_dir=output_path,
        folder_name="",
    )


def binary_to_semantic(yml_input_file: Path) -> bool:
    """
    Convert a binary image stack to a semantic image stack using a YAML configuration file.

    This function reads a YAML file to prepare a semantic image stack from a binary image stack.
    It is designed for a specific use case (NOMAD, AM tensile bars) as of 31 May 2024.

    Parameters
    ----------
    yml_input_file : Path
        The path to the YAML input file containing configuration settings.

    Returns
    -------
    bool
        True if the semantic image stack was successfully created, False otherwise.

    Examples
    --------
    >>> yml_input_file = Path("path/to/config.yml")
    >>> binary_to_semantic(yml_input_file)
    True
    """

    [input_path, output_path, params] = validate_yml(
        yml_input_file, "binary_to_semantic"
    )

    class_labels = binary_with_pores_to_semantic(
        input_path=input_path, output_path=output_path
    )

    print(f"class labels for semantic stack:\n{class_labels}")


def main_binary_to_semantic():
    """
    Run the binary to semantic conversion module from the command line.

    This function serves as the entry point for terminal-based access to the binary to semantic conversion module.
    It is invoked from the command line using the 'binary_to_semantic' command specified in pyproject.toml.
    The function processes a YAML input file to convert a binary image stack to a semantic image stack.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Examples
    --------
    To run the binary to semantic conversion, use the following command in the terminal:
    $ binary_to_semantic path/to/input.yml
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="the .yml user input file")
    args = parser.parse_args()
    yml_input_file = args.input_file

    binary_to_semantic(yml_input_file=yml_input_file)

    print(f"{yml_input_file} processed!")


def main_semantic_to_binary():
    """
    Run the semantic to binary conversion module from the command line.

    This function serves as the entry point for terminal-based access to the semantic to binary conversion module.
    It is invoked from the command line using the 'semantic_to_binary' command specified in pyproject.toml.
    The function processes a YAML input file to convert a semantic image stack to a binary image stack.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Examples
    --------
    To run the semantic to binary conversion, use the following command in the terminal:
    $ semantic_to_binary path/to/input.yml
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="the .yml user input file")
    args = parser.parse_args()
    yml_input_file = args.input_file

    semantic_to_binary(yml_input_file=yml_input_file)

    print(f"{yml_input_file} processed!")


def hdf_to_instance_properties(hdf_path: Path, group_path: str) -> InstanceProperties:
    """
    Read instance analysis data from an HDF5 file and create an InstanceProperties object.

    This function reads data from an HDF5 file that has had instance analysis performed
    with a pre-defined internal structure to create an InstanceProperties object.
    The group name should be the internal 'folder' in the HDF5 file containing the instance
    analysis property data.

    Parameters
    ----------
    hdf_path : Path
        The path to the HDF5 file containing the instance analysis data.
    group_path : str
        The internal 'folder' in the HDF5 file containing the instance analysis property data.

    Returns
    -------
    InstanceProperties
        An InstanceProperties object containing the instance analysis data.

    Examples
    --------
    >>> hdf_path = Path("path/to/instance_analysis.h5")
    >>> group_path = "instance_properties"
    >>> instance_properties = hdf_to_instance_properties(hdf_path, group_path)
    >>> print(instance_properties)
    InstanceProperties(source_name='instance_properties', labels=InstanceLabels(data=array([0, 1, 2, ...])), n_voxels=[NVoxel(value=...), ...], equivalent_sphere_diameters=[Length(value=..., unit=<Units.MICRON: 'micron'>), ...], centroids=Centroids(data=[Centroid(cx=Length(value=..., unit=<Units.MICRON: 'micron'>), cy=Length(value=..., unit=<Units.MICRON: 'micron'>), cz=Length(value=..., unit=<Units.MICRON: 'micron'>)), ...]), ellipsoids=BestFitEllipsoids(data=[BestFitEllipsoid(a=EllipsoidAxis(length=Length(value=..., unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=..., v=..., w=...)), b=EllipsoidAxis(length=Length(value=..., unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=..., v=..., w=...)), c=EllipsoidAxis(length=Length(value=..., unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=..., v=..., w=...))), ...]), surface_areas=EllipsoidSurfaceAreas(data=[Area(value=..., unit_squared=<Units.MICRON: 'micron'>), ...]), volumes=EllipsoidVolumes(data=[Volume(value=..., unit_cubed=<Units.MICRON: 'micron'>), ...]))
    """

    with h5py.File(hdf_path, "r") as f:
        instance_data = f[group_path]
        hdf_n_voxels = np.squeeze(instance_data["num_voxels"][:])
        hdf_equiv_diam = np.squeeze(instance_data["equivalent_sphere_diameters"][:])
        equiv_diam_unit = Units(
            instance_data["equivalent_sphere_diameters"].attrs["units"]
        )
        hdf_centroids = np.squeeze(instance_data["centroids"][:])
        centroid_unit = Units(instance_data["centroids"].attrs["units"])
        hdf_semi_axes = np.squeeze(instance_data["semi-axis_lengths"][:])
        semi_axes_unit = Units(instance_data["semi-axis_lengths"].attrs["units"])
        hdf_vectors = np.squeeze(instance_data["axis_vectors"][:])
        hdf_surface_areas = np.squeeze(instance_data["ellipsoid_surface_areas"][:])
        surface_area_unit = Units(
            instance_data["ellipsoid_surface_areas"].attrs["units_squared"]
        )
        hdf_volumes = np.squeeze(instance_data["ellipsoid_volumes"][:])
        volume_unit = Units(instance_data["ellipsoid_volumes"].attrs["units_cubed"])

    labels = np.arange(0, len(hdf_equiv_diam), dtype=int)
    n_voxels = [NVoxel(value=i) for i in hdf_n_voxels]
    equivalent_sphere_diameters = [
        Length(value=i, unit=equiv_diam_unit) for i in hdf_equiv_diam
    ]
    centroids = Centroids(
        data=[
            Centroid(
                cx=Length(value=i[0], unit=centroid_unit),
                cy=Length(value=i[1], unit=centroid_unit),
                cz=Length(value=i[2], unit=centroid_unit),
            )
            for i in hdf_centroids
        ]
    )

    ellipsoids = BestFitEllipsoids(
        [
            BestFitEllipsoid(
                a=EllipsoidAxis(
                    length=Length(value=i[0], unit=semi_axes_unit),
                    orientation=UnitVector(u=j[0], v=j[1], w=j[2]),
                ),
                b=EllipsoidAxis(
                    length=Length(value=i[1], unit=semi_axes_unit),
                    orientation=UnitVector(u=j[3], v=j[4], w=j[5]),
                ),
                c=EllipsoidAxis(
                    length=Length(value=i[2], unit=semi_axes_unit),
                    orientation=UnitVector(u=j[6], v=j[7], w=j[8]),
                ),
            )
            for i, j in zip(hdf_semi_axes, hdf_vectors)
        ]
    )

    surface_areas = EllipsoidSurfaceAreas(
        [Area(value=i, unit_squared=surface_area_unit) for i in hdf_surface_areas]
    )

    volumes = EllipsoidVolumes(
        [Volume(value=i, unit_cubed=volume_unit) for i in hdf_volumes]
    )

    instance_props = InstanceProperties(
        source_name=group_path,
        labels=InstanceLabels(data=labels),
        n_voxels=n_voxels,
        equivalent_sphere_diameters=equivalent_sphere_diameters,
        centroids=centroids,
        ellipsoids=ellipsoids,
        surface_areas=surface_areas,
        volumes=volumes,
    )

    # n_voxels: list[NVoxel]  # could have a InstanceImageStack
    # equivalent_sphere_diameters: list[Length]
    # centroids: Centroids  # could have a InstanceImageStack
    # ellipsoids: BestFitEllipsoids
    # surface_areas: SurfaceAreas)

    return instance_props


def hdf_to_metadata(hdf_path: Path, dataset_path: str) -> MetaData:
    """
    Extract metadata from an HDF5 dataset containing string metadata attributes.

    This function reads metadata attributes from a specified dataset within an HDF5 file
    and returns a MetaData object.

    Parameters
    ----------
    hdf_path : Path
        The path to the HDF5 file containing the dataset.
    dataset_path : str
        The internal path to the dataset within the HDF5 file.

    Returns
    -------
    MetaData
        A MetaData object containing the extracted metadata.

    Examples
    --------
    >>> hdf_path = Path("path/to/data.h5")
    >>> dataset_path = "dataset"
    >>> metadata = hdf_to_metadata(hdf_path, dataset_path)
    >>> print(metadata)
    MetaData(data_volume=DataVolume(x_width=..., y_height=..., z_image_count=...), resolution=Resolution(dx=Length(value=..., unit=<Units.MICRON: 'micron'>), dy=Length(value=..., unit=<Units.MICRON: 'micron'>), dz=Length(value=..., unit=<Units.MICRON: 'micron'>)), pixel_units=<Units.MICRON: 'micron'>, origin=Origin(x0=Length(value=..., unit=<Units.MICRON: 'micron'>), y0=Length(value=..., unit=<Units.MICRON: 'micron'>), z0=Length(value=..., unit=<Units.MICRON: 'micron'>)))
    """

    with h5py.File(hdf_path, "r") as f:
        dataset = f[dataset_path]

        x_width = int(dataset.attrs["x_width (pixels)"])
        y_height = int(dataset.attrs["y_height (pixels)"])
        z_image_count = int(dataset.attrs["z_image_count (pixels)"])
        data_volume = DataVolume(
            x_width=x_width, y_height=y_height, z_image_count=z_image_count
        )

        pixel_units = Units(dataset.attrs["Resolution, units"])
        dx = Length(value=float(dataset.attrs["Resolution, dx"]), unit=pixel_units)
        dy = Length(value=float(dataset.attrs["Resolution, dy"]), unit=pixel_units)
        dz = Length(value=float(dataset.attrs["Resolution, dz"]), unit=pixel_units)
        resolution = Resolution(
            dx=dx,
            dy=dy,
            dz=dz,
        )

        origin_units = Units(dataset.attrs["Origin, units"])
        x0 = Length(value=float(dataset.attrs["Origin, x0"]), unit=origin_units)
        y0 = Length(value=float(dataset.attrs["Origin, y0"]), unit=origin_units)
        z0 = Length(value=float(dataset.attrs["Origin, z0"]), unit=origin_units)
        origin = Origin(
            x0=x0,
            y0=y0,
            z0=z0,
        )

    metadata = MetaData(
        data_volume=data_volume,
        resolution=resolution,
        pixel_units=pixel_units,
        origin=origin,
    )

    return metadata


def centroids_to_ndarray(centroids: Centroids) -> np.ndarray:
    """
    Convert centroid data type into a NumPy array.

    This function converts a Centroids object into a NumPy array with each row representing
    the (cx, cy, cz) coordinates of a centroid.

    Parameters
    ----------
    centroids : Centroids
        The Centroids object containing the centroid data.

    Returns
    -------
    np.ndarray
        A NumPy array with shape (n, 3), where n is the number of centroids, and each row
        contains the (cx, cy, cz) coordinates of a centroid.

    Examples
    --------
    >>> centroids = Centroids(data=[
    ...     Centroid(cx=Length(0.0, Units.MICRON), cy=Length(0.0, Units.MICRON), cz=Length(0.0, Units.MICRON)),
    ...     Centroid(cx=Length(1.0, Units.MICRON), cy=Length(1.0, Units.MICRON), cz=Length(1.0, Units.MICRON)),
    ...     Centroid(cx=Length(2.0, Units.MICRON), cy=Length(2.0, Units.MICRON), cz=Length(2.0, Units.MICRON))
    ... ])
    >>> centroids_to_ndarray(centroids)
    array([[0., 0., 0.],
           [1., 1., 1.],
           [2., 2., 2.]])
    """

    data = centroids.data
    cx = [i.cx.value for i in data]
    cy = [i.cy.value for i in data]
    cz = [i.cz.value for i in data]

    ndarray = np.array((cx, cy, cz), dtype=float).T

    return ndarray


def ellipsoids_to_ndarray(
    ellipsoids: BestFitEllipsoids,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Convert ellipsoid data type into NumPy arrays.

    This function converts a BestFitEllipsoids object into two NumPy arrays:
    one for the axis lengths and one for the axis vectors.

    Parameters
    ----------
    ellipsoids : BestFitEllipsoids
        The BestFitEllipsoids object containing the ellipsoid data.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        A tuple containing:
        - A NumPy array with shape (n, 3) for the axis lengths, where n is the number of ellipsoids.
        - A NumPy array with shape (n, 9) for the axis vectors, where n is the number of ellipsoids.

    Examples
    --------
    >>> ellipsoids = BestFitEllipsoids(data=[
    ...     BestFitEllipsoid(
    ...         a=EllipsoidAxis(length=Length(5.0, Units.MICRON), orientation=UnitVector(u=1.0, v=0.0, w=0.0)),
    ...         b=EllipsoidAxis(length=Length(3.0, Units.MICRON), orientation=UnitVector(u=0.0, v=1.0, w=0.0)),
    ...         c=EllipsoidAxis(length=Length(2.0, Units.MICRON), orientation=UnitVector(u=0.0, v=0.0, w=1.0))
    ...     ),
    ...     BestFitEllipsoid(
    ...         a=EllipsoidAxis(length=Length(6.0, Units.MICRON), orientation=UnitVector(u=0.707, v=0.707, w=0.0)),
    ...         b=EllipsoidAxis(length=Length(4.0, Units.MICRON), orientation=UnitVector(u=-0.707, v=0.707, w=0.0)),
    ...         c=EllipsoidAxis(length=Length(3.0, Units.MICRON), orientation=UnitVector(u=0.0, v=0.0, w=1.0))
    ...     )
    ... ])
    >>> axis_lengths, axis_vectors = ellipsoids_to_ndarray(ellipsoids)
    >>> print(axis_lengths)
    array([[5., 3., 2.],
           [6., 4., 3.]])
    >>> print(axis_vectors)
    array([[ 1.   ,  0.   ,  0.   ,  0.   ,  1.   ,  0.   ,  0.   ,  0.   ,  1.   ],
           [ 0.707,  0.707,  0.   , -0.707,  0.707,  0.   ,  0.   ,  0.   ,  1.   ]])
    """

    data = ellipsoids.data
    # axis lengths:
    a = [i.a.length.value for i in data]
    b = [i.b.length.value for i in data]
    c = [i.c.length.value for i in data]
    axis_lengths = np.array((a, b, c), dtype=float).T

    # axes vectors
    a_u = [i.a.orientation.u for i in data]
    a_v = [i.a.orientation.v for i in data]
    a_w = [i.a.orientation.w for i in data]

    b_u = [i.b.orientation.u for i in data]
    b_v = [i.b.orientation.v for i in data]
    b_w = [i.b.orientation.w for i in data]

    c_u = [i.c.orientation.u for i in data]
    c_v = [i.c.orientation.v for i in data]
    c_w = [i.c.orientation.w for i in data]

    axis_vectors = np.array(
        (a_u, a_v, a_w, b_u, b_v, b_w, c_u, c_v, c_w), dtype=float
    ).T

    return axis_lengths, axis_vectors


def surface_areas_to_ndarray(surface_areas: EllipsoidSurfaceAreas) -> np.ndarray:
    """
    Convert surface area data type into a NumPy array.

    This function converts an EllipsoidSurfaceAreas object into a NumPy array
    containing the surface area values.

    Parameters
    ----------
    surface_areas : EllipsoidSurfaceAreas
        The EllipsoidSurfaceAreas object containing the surface area data.

    Returns
    -------
    np.ndarray
        A NumPy array containing the surface area values.

    Examples
    --------
    >>> surface_areas = EllipsoidSurfaceAreas(data=[
    ...     Area(value=100.0, unit_squared=Units.MICRON),
    ...     Area(value=200.0, unit_squared=Units.MICRON),
    ...     Area(value=300.0, unit_squared=Units.MICRON)
    ... ])
    >>> surface_areas_to_ndarray(surface_areas)
    array([100., 200., 300.])
    """

    data = surface_areas.data
    areas = [i.value for i in data]

    ndarray = np.array((areas), dtype=float).T

    return ndarray


def volumes_to_ndarray(ellipsoid_volumes: EllipsoidVolumes) -> np.ndarray:
    """
    Convert ellipsoid volume data type into a NumPy array.

    This function converts an EllipsoidVolumes object into a NumPy array
    containing the volume values.

    Parameters
    ----------
    ellipsoid_volumes : EllipsoidVolumes
        The EllipsoidVolumes object containing the volume data.

    Returns
    -------
    np.ndarray
        A NumPy array containing the volume values.

    Examples
    --------
    >>> ellipsoid_volumes = EllipsoidVolumes(data=[
    ...     Volume(value=500.0, unit_cubed=Units.MICRON),
    ...     Volume(value=1000.0, unit_cubed=Units.MICRON),
    ...     Volume(value=1500.0, unit_cubed=Units.MICRON)
    ... ])
    >>> volumes_to_ndarray(ellipsoid_volumes)
    array([ 500., 1000., 1500.])
    """

    data = ellipsoid_volumes.data
    volumes = [i.value for i in data]

    ndarray = np.array((volumes), dtype=float).T

    return ndarray


def rmdir(directory: Path) -> None:
    """
    Recursively delete a directory and all its contents.
    (credit to: https://stackoverflow.com/questions/13118029/deleting-folders-in-python-recursively/49782093#49782093)


    This function deletes the specified directory and all its contents, including
    subdirectories and files. If the directory does not exist, the function does nothing.

    Parameters
    ----------
    directory : Path
        The path to the directory to be deleted.

    Returns
    -------
    None

    Examples
    --------
    >>> from pathlib import Path
    >>> directory = Path("path/to/directory")
    >>> rmdir(directory)
    """

    if not directory.exists():
        return

    for item in directory.iterdir():
        if item.is_dir():
            rmdir(item)
        else:
            item.unlink()
    directory.rmdir()


def compare_files(file1: Path, file2: Path, ignore_words: list[str]) -> bool:
    """
    Compare two files line-by-line, ignoring lines that contain specified words.

    This function compares two files line-by-line and ignores lines that contain
    any of the words specified in the `ignore_words` list. If the files are identical
    up to the allowed differences, the function returns True. Otherwise, it returns False.

    Parameters
    ----------
    file1 : Path
        Path to the first file for comparison.
    file2 : Path
        Path to the second file for comparison.
    ignore_words : list of str
        List of words that cause a particular line to be ignored during comparison.
        Use an empty list for strict word-by-word comparison. Use a non-empty list,
        e.g., ["date", "time"], to ignore lines with these words, as they may differ
        from file to file.

    Returns
    -------
    bool
        True if the two files are the same (up to any allowed differences in the
        ignore_words list), False if the two files are different.

    Examples
    --------
    >>> file1 = Path("path/to/file1.txt")
    >>> file2 = Path("path/to/file2.txt")
    >>> ignore_words = ["date", "time"]
    >>> compare_files(file1, file2, ignore_words)
    True
    """

    with open(file=file1, mode="r", encoding="utf-8") as fa:
        with open(file=file2, mode="r", encoding="utf-8") as fb:
            dba = fa.readlines()
            dbb = fb.readlines()

            if len(dba) == len(dbb):
                for k, line in enumerate(dba):
                    if line != dbb[k]:
                        # if "Autogenerated" in line:
                        iw_in_fa_line = [iw in line for iw in ignore_words]
                        iw_in_fb_line = [iw in dbb[k] for iw in ignore_words]
                        in_either_line = iw_in_fa_line + iw_in_fb_line
                        # if any(iw in [line + dbb[k]] for iw in ignore_words):
                        if any(in_either_line):
                            # skip comparison of the ignore_words case(s)
                            print("Skip comparison of word in ignore_word.")
                        else:
                            print(f"Files differ at line {k}:")
                            print(f"\t file1: {line}")
                            print(f"\t file2: {dbb[k]}")
                            return False

            else:
                print("Files have non-equal line length.")
                print("Line-by-line comparison not performed.")
                return False

    return True  # files are the same


def date_time_utc() -> str:
    """
    Returns the current date and time in ISO format in the UTC time zone.

    This function returns the current date and time in ISO 8601 format, with the time zone
    specified as UTC. The colons (:) and periods (.) in the time string are replaced with
    underscores (_) for compatibility with file naming conventions.

    Returns
    -------
    str
        The current date and time in ISO format in the UTC time zone.

    Examples
    --------
    >>> date_time_utc()
    '2024-08-13_UTC_12_34_56_789012'
    """

    # ts = datetime.utcnow().isoformat()  # The "naive object in UTC", utcnow is deprecated
    ts = datetime.now(datetime.UTC).isocalendar()  # The "naive object in UTC"
    ts = ts.replace(":", "_").replace(".", "_")  # Overwrite ":", "." with "_"
    ts = ts.replace("T", "_UTC_")  # Overwrite T with UTC time zone indication
    return ts


def underline() -> str:
    """
    Return a CTH commented underline composed of 78 dashes ('-') and a newline character ('\\n').

    This function generates a string that represents a commented underline for CTH (a computational tool),
    consisting of 78 dashes prefixed with a comment character ('* ') and followed by a newline character.

    Returns
    -------
    str
        A string representing a CTH commented underline.

    Examples
    --------
    >>> underline()
    '* ------------------------------------------------------------------------------\\n'
    """

    return "* " + "".join(repeat("-", 78)) + "\n"


def in_a_but_not_in_b(*, a: Iterable, b: Iterable) -> Iterable:
    """
    Return items in Iterable `a` that are not in Iterable `b`.

    This function takes two iterables `a` and `b`, and returns a tuple containing
    all the items that are present in `a` but not in `b`.

    Parameters
    ----------
    a : Iterable
        The first iterable to compare.
    b : Iterable
        The second iterable to compare.

    Returns
    -------
    Tuple
        A tuple containing items that are in `a` but not in `b`.

    Examples
    --------
    >>> a = [1, 2, 3, 4]
    >>> b = [3, 4, 5, 6]
    >>> in_a_but_not_in_b(a=a, b=b)
    (1, 2)
    """

    result = ()  # empty tuple
    for item in a:
        if item not in b:
            result = result + (item,)

    return result


def pairwise(x: Iterable) -> Iterable:
    """
    Return successive overlapping pairs taken from the input iterable.
    This appears to be implemented in Python 3.10
    https://docs.python.org/3/library/itertools.html#itertools.pairwise
    but we currently use 3.9, so we implement `pairwise` here.

    The number of 2-tuples in the output iterator will be one fewer than the
    number of inputs. It will be empty if the input iterable has fewer than
    two values.

    Parameters
    ----------
    x : Iterable
        The input iterable from which to generate pairs.

    Returns
    -------
    Iterable[Tuple]
        An iterable of 2-tuples containing successive overlapping pairs from the input iterable.

    Examples
    --------
    >>> list(pairwise('ABCDEFG'))
    [('A', 'B'), ('B', 'C'), ('C', 'D'), ('D', 'E'), ('E', 'F'), ('F', 'G')]
    >>> list(pairwise([1, 2, 3, 4]))
    [(1, 2), (2, 3), (3, 4)]
    """

    a, b = tee(x)
    next(b, None)
    return zip(a, b)


def pairwise_circular(x: Iterable) -> Iterable:
    """
    Return successive overlapping pairs taken from the input iterable.

    The number of 2-tuples in the output iterator will be one fewer than the
    number of inputs. It will be empty if the input iterable has fewer than
    two values.

    Parameters
    ----------
    x : Iterable
        The input iterable from which to generate pairs.

    Returns
    -------
    Iterable[Tuple]
        An iterable of 2-tuples containing successive overlapping pairs from the input iterable.

    Examples
    --------
    >>> list(pairwise('ABCDEFG'))
    [('A', 'B'), ('B', 'C'), ('C', 'D'), ('D', 'E'), ('E', 'F'), ('F', 'G')]
    >>> list(pairwise([1, 2, 3, 4]))
    [(1, 2), (2, 3), (3, 4)]
    """

    a = cycle(x)
    next(a)
    return zip(x, a)


def metadata_to_dict(metadata: MetaData) -> dict:
    """
    Convert MetaData to a dictionary.

    This function converts a MetaData object to a dictionary representation.

    Parameters
    ----------
    metadata : MetaData
        The metadata of the Data Volume.

    Returns
    -------
    dict
        A dictionary containing the metadata.

    Examples
    --------
    >>> metadata = MetaData(
    ...     data_volume=DataVolume(x_width=256, y_height=256, z_image_count=10),
    ...     resolution=Resolution(
    ...         dx=Length(1.0, Units.MICRON),
    ...         dy=Length(1.0, Units.MICRON),
    ...         dz=Length(1.0, Units.MICRON)
    ...     ),
    ...     pixel_units=Units.MICRON,
    ...     origin=Origin(
    ...         x0=Length(0.0, Units.MICRON),
    ...         y0=Length(0.0, Units.MICRON),
    ...         z0=Length(0.0, Units.MICRON)
    ...     )
    ... )
    >>> metadata_to_dict(metadata)
    {
        'x_width (pixels)': '256',
        'y_height (pixels)': '256',
        'z_image_count (pixels)': '10',
        'Resolution, dx': '1.0',
        'Resolution, dy': '1.0',
        'Resolution, dz': '1.0',
        'Resolution, units': 'micron',
        'Origin, x0': '0.0',
        'Origin, y0': '0.0',
        'Origin, z0': '0.0',
        'Origin, units': 'micron'
    }
    """

    data_volume = metadata.data_volume
    resolution = metadata.resolution
    # pixel_units = metadata.pixel_units
    origin = metadata.origin

    x_width = data_volume.x_width
    y_height = data_volume.y_height
    z_image_count = data_volume.z_image_count

    dx = resolution.dx
    dy = resolution.dz
    dz = resolution.dz

    meta_dict = {
        "x_width (pixels)": str(x_width),
        "y_height (pixels)": str(y_height),
        "z_image_count (pixels)": str(z_image_count),
        "Resolution, dx": str(dx.value),
        "Resolution, dy": str(dy.value),
        "Resolution, dz": str(dz.value),
        "Resolution, units": str(dx.unit.value),
        "Origin, x0": str(origin.x0.value),
        "Origin, y0": str(origin.y0.value),
        "Origin, z0": str(origin.z0.value),
        "Origin, units": str(origin.x0.unit.value),
    }
    # print(meta_dict)

    return meta_dict


def yaml_to_dict(path_file_input: Path) -> dict:
    """
    Convert a YAML file to a dictionary.

    This function reads a YAML file and converts its contents to a dictionary.

    Parameters
    ----------
    path_file_input : Path
        The full path to the input YAML file.

    Returns
    -------
    dict
        A dictionary containing the contents of the YAML file.

    Raises
    ------
    TypeError
        If the file type is not supported.
    OSError
        If there is an error opening or decoding the YAML file.

    Examples
    --------
    >>> path_file_input = Path("path/to/config.yml")
    >>> yaml_to_dict(path_file_input)
    {'key1': 'value1', 'key2': 'value2', ...}
    """

    file_type = Path(path_file_input).suffix.casefold()
    supported_types = (".yaml", ".yml")

    if file_type not in supported_types:
        raise TypeError("Only file types .yaml, and .yml are supported.")

    try:
        with open(file=path_file_input, mode="r", encoding="utf-8") as stream:
            db = yaml.load(stream, Loader=yaml.SafeLoader)  # Load YAML file
    except yaml.YAMLError as error:
        print(f"Error with YAML file: {error}")
        print(f"Could not open or decode: {path_file_input}")
        raise OSError from error

    print(f"Success: database created from file: {path_file_input}")
    print("key, value, type")
    print("---, -----, ----")
    for key, value in db.items():
        print(f"{key}, {value}, {type(value)}")

    return db


def dict_to_yaml(db: dict, file: str) -> Path:
    """
    Convert a dictionary to a YAML file.

    This function writes the contents of a dictionary to a YAML file.

    Parameters
    ----------
    db : dict
        The dictionary to be converted to YAML.
    file : str
        The path to the output YAML file.

    Returns
    -------
    Path
        The path to the created YAML file.

    Examples
    --------
    >>> db = {'key1': 'value1', 'key2': 'value2'}
    >>> file = "path/to/output.yml"
    >>> dict_to_yaml(db, file)
    PosixPath('path/to/output.yml')
    """

    with open(file, "w", encoding="utf-8") as out_file:
        yaml.dump(db, out_file, default_flow_style=False)  # Write dictionary to YAML

    return Path(file)


def ndarray_to_img(
    *,
    data: np.ndarray,
    slice_axis: CartesianAxis3D,
    parent_dir: Path,
    folder_name: str,
    pad_length: int = 4,
    file_type: str = ".tif",
) -> bool:
    """
    Convert an ndarray to an image stack and save it to a specified directory.

    This function takes a NumPy ndarray and creates an image stack, saving the images
    into a user-specified directory.

    Parameters
    ----------
    data : np.ndarray
        The semantic labels.
    slice_axis : CartesianAxis3D
        The axis along which to slice the ndarray:
        - 0 for Z axis
        - 1 for Y axis
        - 2 for X axis
    parent_dir : Path
        The parent directory to save the image folder.
    folder_name : str
        The folder name to save images.
    pad_length : int, optional
        The number of digits to pad the file names with (default is 4).
    file_type : str, optional
        The image file type (default is ".tif").

    Returns
    -------
    bool
        True if the images were successfully created.

    Examples
    --------
    >>> data = np.random.randint(0, 255, (10, 256, 256), dtype=np.uint8)
    >>> slice_axis = CartesianAxis3D.Z
    >>> parent_dir = Path("path/to/save")
    >>> folder_name = "image_stack"
    >>> ndarray_to_img(data=data, slice_axis=slice_axis, parent_dir=parent_dir, folder_name=folder_name)
    True
    """

    img_dir = parent_dir.joinpath(folder_name).expanduser()
    img_dir.mkdir(parents=True, exist_ok=True)
    n_slices = data.shape[slice_axis.value]

    for i in range(n_slices):
        fname = f"{img_dir}/{i:0{pad_length}}{file_type}"
        mode = "L" if data.dtype == np.int8 else None

        match slice_axis:
            case CartesianAxis3D.Z:
                img = Image.fromarray(data[i, :, :], mode=mode)
            case CartesianAxis3D.Y:
                img = Image.fromarray(data[:, i, :], mode=mode)
            case CartesianAxis3D.X:
                img = Image.fromarray(data[:, :, i], mode=mode)
            case _:
                raise ValueError(
                    f"Unknown slice_axis value {slice_axis}, value must be 0, 1, or 2."
                )

        img.save(fname)
    return True


def read_images(
    file_dir: Path,
    file_type: str = ".tif",
) -> np.ndarray:
    """
    Read images from a directory and return a NumPy array representation of the images.

    This function reads images from the specified directory and returns a NumPy array
    representation of the images.

    Parameters
    ----------
    file_dir : Path
        The fully pathed location of the images.
    file_type : str, optional
        The image type (default is ".tif").

    Returns
    -------
    np.ndarray
        A NumPy array representation of the images.

    Raises
    ------
    FileNotFoundError
        If no images of the specified type are found in the directory.

    Examples
    --------
    >>> file_dir = Path("path/to/images")
    >>> read_images(file_dir, file_type=".tif")
    array([[[...], [...], ...], [[...], [...], ...], ...])
    """

    image_list = list(glob.glob(f"{str(file_dir.as_posix())}/*{file_type}"))

    if len(image_list) == 0:
        raise FileNotFoundError(
            f"File type of {file_type} not found in directory: {str(file_dir)}"
        )

    image_list.sort()  # Sort images in ascending order

    image_stack = np.array([np.array(Image.open(f)) for f in image_list])

    # Handle the case where only a single image is read
    if image_stack.ndim < 3:
        image_stack = np.expand_dims(image_stack, axis=-1)
        print(f"Only single image read, new image array size: {image_stack.shape}")
    else:
        print(f"Images read, image array size: {image_stack.shape}")

    return image_stack



"""This module tests individual functions for the core module.

Example:
    To run
    - activate the virutal environment, then
    $ cd ~/recond3d

    pytest (without code coverage)
    pytest recon3d/tests/test_instance_analysis.py -v  # -v is for verbose

    to run just a single test in this module, for example
    pytest recon3d/tests/test_instance_analysis.py::test_read_images -v

"""

# Python standard libarires
import glob
from pathlib import Path

# Third party libraries
import numpy as np
import pytest

# local libraries
# import recon3d.feature_analysis as fa
import recon3d.instance_analysis as ia
import recon3d.utility as ut
import recon3d.constants as cs
from recon3d.types import *

from recon3d.static_test_paths import *


def create_instance_stack(instance_name):
    """Generate object required for further tests"""

    db = ut.yaml_to_dict(INSTANCE_ANALYSIS_YML)

    # overwrite so that this test runs on any machine
    db["semantic_image_dir"] = str(
        Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")
    )
    temp_path = Path(__file__).parent.joinpath("data", "temp.yml")

    # make sure the recon3d/examples/temp.yml doesn't already
    # exist from a previous test run or from a test paused
    # midway through
    try:
        temp_path.unlink()  # delete the temp.yml if it exists aready
    except FileNotFoundError:
        pass

    # make temp yml
    bb = ut.dict_to_yaml(db, str(temp_path))

    found_stack = ia.process_image_stack(yml_input_file=bb)
    print(db)
    # instance_name = "pore"
    instance_value = db["class_labels"][instance_name]["value"]
    min_feat_size = db["class_labels"][instance_name]["instance_analysis"][
        "min_feature_size"
    ]

    found_instance = ia.semantic_to_instance(
        semantic_stack=found_stack,
        instance_name=instance_name,
        instance_value=instance_value,
        min_feature_size=min_feat_size,
    )

    temp_path.unlink()

    return found_instance


# @pytest.fixture
# def test_file() -> Path:
#     """The relative path and file string locating the default yml test file."""

#     return Path(__file__).parent.joinpath("examples", "instance_analysis.yml")


def test_process_image_stack():
    """Test that a yml input file that specifies an image stack can be read in
    and used to create a h5 file.
    """

    aa = INSTANCE_ANALYSIS_YML
    db = ut.yaml_to_dict(aa)
    # overwrite so that this test runs on any machine
    db["semantic_image_dir"] = str(
        Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")
    )

    temp_path = Path(__file__).parent.joinpath("data", "temp.yml")

    try:
        temp_path.unlink()  # delete the temp.yml if it exists aready
    except FileNotFoundError:
        pass

    bb = ut.dict_to_yaml(db, str(temp_path))

    found_stack = ia.process_image_stack(yml_input_file=bb)

    images_path = Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")

    known_stack = SemanticImageStack(
        name="semantic_output",
        data=ut.read_images(images_path, "tif"),
        metadata=MetaData(
            data_volume=DataVolume(
                x_width=923,
                y_height=925,
                z_image_count=21,
            ),
            resolution=Resolution(
                dx=Length(1, Units.MICRON),
                dy=Length(1, Units.MICRON),
                dz=Length(1, Units.MICRON),
            ),
            pixel_units="micron",
            origin=Origin(
                x0=Length(0.0, Units.MICRON),
                y0=Length(0.0, Units.MICRON),
                z0=Length(0.0, Units.MICRON),
            ),
        ),
    )

    assert found_stack.data.all() == known_stack.data.all()

    temp_path.unlink()


def test_read_images():
    """Tests that either images exist and can be read, or that None is returned."""

    dataset_path = Path(__file__).parent.joinpath("data", "cylinder_machined_semantic")
    image_type = ".tif"

    output = ut.read_images(file_dir=dataset_path, file_type=image_type)

    assert isinstance(output, np.ndarray)

    # now specify some non-existed files
    image_type = (
        ".gif"  # overwrite, this file type does not exist in the dataset_path folder.
    )
    with pytest.raises(FileNotFoundError):
        output = ut.read_images(file_dir=dataset_path, file_type=image_type)


def test_semantic_to_instance():
    """Test the instance image stack is properly extracted from the semantic image stack."""

    found_instance = create_instance_stack("pore")

    assert found_instance.nlabels == 103


def test_instance_mask():
    """Test the instance mask is properly created"""
    instance_stack = create_instance_stack("pore")
    instance_ids = ia.instance_indices(instance_stack)
    indices = instance_ids.indices
    labels = instance_ids.labels.data
    found_idxs = indices[1]

    known_idxs = np.array([[1, 14, 52], [1, 13, 52]])
    assert len(labels) == len(np.unique(instance_stack.data))
    assert found_idxs == pytest.approx(known_idxs)


def test_centroid():
    """Test the centroid is properly calcualted"""
    instance_stack = create_instance_stack("pore")
    instance_ids = ia.instance_indices(instance_stack)
    indices = instance_ids.indices
    found_idxs = indices[1]
    resolution = Resolution(
        dx=Length(2.0, unit=Units.MICRON),
        dy=Length(2.0, unit=Units.MICRON),
        dz=Length(2.0, unit=Units.MICRON),
    )
    origin = Origin(
        x0=Length(1.0, unit=Units.MICRON),
        y0=Length(1.0, unit=Units.MICRON),
        z0=Length(1.0, unit=Units.MICRON),
    )
    found_centroid = ia.center_of_mass(
        indices=found_idxs, resolution=resolution, origin=origin
    )

    known_centroid = Centroid(
        cx=Length((52 * 2.0) + 1.0, unit=Units.MICRON),
        cy=Length((13.5 * 2.0) + 1.0, unit=Units.MICRON),
        cz=Length((1 * 2.0) + 1.0, unit=Units.MICRON),
    )
    assert found_centroid == pytest.approx(known_centroid)


def test_centroid_pix():
    "test the centroid in pixels/voxels units is properly calculated"
    centroid = Centroid(
        cx=Length(5.2, unit=Units.MICRON),
        cy=Length(1.1, unit=Units.MICRON),
        cz=Length(3.5, unit=Units.MICRON),
    )
    resolution = Resolution(
        dx=Length(2.0, unit=Units.MICRON),
        dy=Length(2.0, unit=Units.MICRON),
        dz=Length(2.0, unit=Units.MICRON),
    )
    origin = Origin(
        x0=Length(1.0, unit=Units.MICRON),
        y0=Length(1.0, unit=Units.MICRON),
        z0=Length(1.0, unit=Units.MICRON),
    )
    found_centroid_px = ia.centroid_pix(
        centroid=centroid, resolution=resolution, origin=origin
    )

    known_centroid_px = Centroid(
        cx=Length(2.1, Units.VOXEL),
        cy=Length(0.05, Units.VOXEL),
        cz=Length(1.25, Units.VOXEL),
    )

    assert found_centroid_px.cx.value == pytest.approx(known_centroid_px.cx.value)
    assert found_centroid_px.cy.value == pytest.approx(known_centroid_px.cy.value)
    assert found_centroid_px.cz.value == pytest.approx(known_centroid_px.cz.value)


def test_calc_moment():
    "test the calc_moment properly calculated"
    instance_stack = create_instance_stack("pore")
    instance_ids = ia.instance_indices(instance_stack)
    indices = instance_ids.indices
    found_idxs = indices[73]

    resolution = Resolution(
        dx=Length(2.0, Units.MICRON),
        dy=Length(2.0, Units.MICRON),
        dz=Length(2.0, Units.MICRON),
    )
    origin = Origin(
        x0=Length(1.0, Units.MICRON),
        y0=Length(1.0, Units.MICRON),
        z0=Length(1.0, Units.MICRON),
    )
    found_centroid = ia.center_of_mass(
        indices=found_idxs, resolution=resolution, origin=origin
    )
    found_centroid_px = ia.centroid_pix(
        centroid=found_centroid, resolution=resolution, origin=origin
    )
    found_moment = ia.calc_moment(
        indices=found_idxs, p=2, q=2, r=2, centroid_px=found_centroid_px
    )
    known_moment = (0.5**2) ** 3
    assert found_moment == pytest.approx(known_moment)


def test_fit_ellipsoid():
    "test the fit_ellipsoid properly calculated"
    instance_stack = create_instance_stack("pore")
    instance_ids = ia.instance_indices(instance_stack)
    indices = instance_ids.indices
    found_idxs = indices[1]
    resolution = Resolution(
        dx=Length(2.0, unit=Units.MICRON),
        dy=Length(2.0, unit=Units.MICRON),
        dz=Length(2.0, unit=Units.MICRON),
    )
    origin = Origin(
        x0=Length(1.0, unit=Units.MICRON),
        y0=Length(1.0, unit=Units.MICRON),
        z0=Length(1.0, unit=Units.MICRON),
    )
    found_centroid = ia.center_of_mass(
        indices=found_idxs, resolution=resolution, origin=origin
    )

    found_ellipsoid = ia.fit_ellipsoid(
        indices=found_idxs,
        centroid=found_centroid,
        resolution=resolution,
        origin=origin,
    )

    a_vector = UnitVector(u=0.0, v=1.0, w=0.0)
    a = EllipsoidAxis(
        length=Length(2.23607, unit=Units.MICRON),
        orientation=a_vector,
    )
    b_vector = UnitVector(
        u=0.0,
        v=0.0,
        w=1.0,
    )
    b = EllipsoidAxis(
        length=Length(0.0, unit=Units.MICRON),
        orientation=b_vector,
    )
    c_vector = UnitVector(
        u=1.0,
        v=0.0,
        w=0.0,
    )
    c = EllipsoidAxis(
        length=Length(0.0, unit=Units.MICRON),
        orientation=c_vector,
    )
    known_ellipsoid = BestFitEllipsoid(a=a, b=b, c=c)

    # TODO look into __repr__ method to directly compare BestFitEllipsoid obejcts
    assert found_ellipsoid.a.length.value == pytest.approx(
        known_ellipsoid.a.length.value
    )
    assert found_ellipsoid.b.length.value == pytest.approx(
        known_ellipsoid.b.length.value
    )
    assert found_ellipsoid.c.length.value == pytest.approx(
        known_ellipsoid.c.length.value
    )


def test_equivalent_spherical_diameter():
    """test calculation of equivalent diameter"""
    num_vox = 4
    res = Resolution(
        dx=Length(2.0, unit=Units.MICRON),
        dy=Length(2.0, unit=Units.MICRON),
        dz=Length(2.0, unit=Units.MICRON),
    )
    known_diam = 3.93898
    found_diam = ia.equivalent_spherical_diameter(
        n_voxels=num_vox,
        resolution=res,
    )
    assert known_diam == pytest.approx(found_diam.value)

    num_vox2 = 27
    res2 = Resolution(
        dx=Length(1.1, unit=Units.MICRON),
        dy=Length(1.1, unit=Units.MICRON),
        dz=Length(2.0, unit=Units.MICRON),
    )
    known_diam2 = 4.9972
    found_diam2 = ia.equivalent_spherical_diameter(
        n_voxels=num_vox2,
        resolution=res2,
    )
    assert known_diam2 == pytest.approx(found_diam2.value)


def test_num_voxels():
    """test calculation of number of voxels of instance IDs"""
    instance_stack = create_instance_stack("pore")
    instance_ids = ia.instance_indices(instance_stack)
    n_voxels = ia.num_voxels(indices=instance_ids)
    data_vol = instance_stack.metadata.data_volume
    total_voxels = data_vol.x_width * data_vol.y_height * data_vol.z_image_count

    found_n_vox = [i.value for i in n_voxels]
    assert sum(found_n_vox) == total_voxels
    assert n_voxels[3].value == 4


def test_minimum_size_filter():
    """test application of minimum size filter"""
    instance_stack = create_instance_stack("pore")
    instance_ids = ia.instance_indices(instance_stack)
    filtered_stack, filtered_ids = ia.minimum_size_filter(
        instance_stack,
        instance_ids,
    )
    assert len(filtered_ids.labels.data) == 6
    assert filtered_stack.nlabels == 5
    assert len(filtered_ids.indices[1]) == 21

    data_vol = filtered_stack.metadata.data_volume
    known_voxels = data_vol.x_width * data_vol.y_height * data_vol.z_image_count

    idx = filtered_ids.indices

    # list comprehension on list of arrays doesn't seem to work
    # found_voxels = [len(idx[i]) for i in idx]

    found_voxels = 0
    for i in idx:
        found_voxels += len(i)
    assert known_voxels == found_voxels


def test_instance_analysis_include():
    """test whether instance analysis should be ran"""
    aa = INSTANCE_ANALYSIS_YML
    db = ut.yaml_to_dict(aa)
    bb = ia.instance_analysis_included(settings=db, key="air")
    assert bb is False

    cc = ia.instance_analysis_included(settings=db, key="metal")
    assert cc is False

    dd = ia.instance_analysis_included(settings=db, key="pore")
    assert dd is True

    ee = ia.instance_analysis_included(settings=db, key="rand")
    assert ee is False


def test_instance_properties():
    instance_stack = create_instance_stack("pore")
    instance_ids = ia.instance_indices(instance_stack)

    found_properties = ia.instance_properties(
        instance_stack=instance_stack, inst_indices=instance_ids
    )
    assert found_properties.source_name == "pore"
    assert len(found_properties.labels.data) == 104
    assert found_properties.n_voxels[55].value == 4
    assert found_properties.ellipsoids.data[53].a.length.value == pytest.approx(
        1.118034
    )

    known_centroid = Centroid(
        cx=Length(12.5, Units.MICRON),
        cy=Length(81.5, Units.MICRON),
        cz=Length(14.0, Units.MICRON),
    )
    found_centroid = found_properties.centroids.data[68]

    assert found_centroid == known_centroid

    # rod_demo_file = Path(__file__).parent.joinpath(
    #     "examples", "rods_instance_analysis.yml"
    # )
    # instance_stack = create_instance_stack(rod_demo_file, "rods")
    # instance_ids = ia.instance_indices(instance_stack)
    # found_properties = ia.instance_properties(
    #     instance_stack=instance_stack, inst_indices=instance_ids
    # )

    # assert found_properties.ellipsoids.data[4].a.orientation.u == pytest.approx(
    #     -0.584283594756
    # )


def test_nth_nearest_neighbor_distance():
    """test calculation of nth nearest neighbors"""
    instance_stack = create_instance_stack("pore")
    instance_indices = ia.instance_indices(instance_stack)
    instance_props = ia.instance_properties(
        instance_stack=instance_stack,
        inst_indices=instance_indices,
    )

    nearest_neighbors = ia.nearest_neighbor_distance(centroids=instance_props.centroids)

    assert nearest_neighbors.instance_id[7] == 12
    assert nearest_neighbors.distances[7].value == pytest.approx(6.576473218982953)


def test_ellipsoid_surface_area():
    """test calculation of ellipsoid surface area"""
    ellipsoid = BestFitEllipsoid(
        a=EllipsoidAxis(
            length=Length(value=2, unit=Units.METER),
            orientation=UnitVector(u=1.0, v=0.0, w=0.0),
        ),
        b=EllipsoidAxis(
            length=Length(value=3, unit=Units.METER),
            orientation=UnitVector(u=0.0, v=1.0, w=0.0),
        ),
        c=EllipsoidAxis(
            length=Length(value=4, unit=Units.METER),
            orientation=UnitVector(u=0.0, v=0.0, w=1.0),
        ),
    )

    found_surface_area = ia.ellipsoid_surface_area(ellipsoid=ellipsoid)
    known_surface_area = 111.604
    assert found_surface_area.value == pytest.approx(known_surface_area)
    assert found_surface_area.unit_squared == Units.METER



"""
This module contains functions related to the `downscale` command line argument 
with a provided input file.

Functions
---------
padded_size(img_stack_dim_size, target_res, original_res, tolerance, limit_factor)
    Determine the expected pad size along a dimension.

pad_amount(img_stack_dim_size, target_res, original_res, tolerance, limit_factor)
    Determine the amount of padding to add along a dimension.

apply_bbox(image_stack, threshold)
    Crop the image stack to the smallest bounding box over the threshold.

bbox_range(image_stack, threshold)
    Calculate the bounding box range in all dimensions.

save_downscale_stack(image_stack, path, folder_suffix)
    Save the new stack as a tiff image stack.

downscale(path_file_input)
    Downscale the image stack based on the provided input file.

main()
    Runs the module from the command line, invoked from pyproject.toml with 'downscale' command.
"""

import argparse
import itertools
import math
from pathlib import Path
from typing import Tuple

import numpy as np
from scipy import ndimage
from pyevtk.hl import gridToVTK

from recon3d.types import CartesianAxis3D
import recon3d.utility as ut


def apply_bbox(image_stack: np.ndarray, min_threshold: float) -> np.ndarray:
    """
    Crop the image stack to the smallest bounding box over the threshold.

    Parameters
    ----------
    image_stack : np.ndarray
        The stack of images as a data cube.
    min_threshold : float
        The minimum threshold value for cropping (values greater than this threshold will be retained).

    Returns
    -------
    np.ndarray
        The cropped image stack.

    Examples
    --------
    >>> image_stack = np.random.rand(10, 10, 10)
    >>> cropped_stack = apply_bbox(image_stack, 0.5)
    Cropping image stack to bounding box...
    """
    print("Cropping image stack to bounding box...")

    (z_start, z_end, y_start, y_end, x_start, x_end) = bbox_range(
        image_stack, min_threshold
    )

    # Slice the image stack to the bbox array
    # Add one to include the full range of the bounding box because of the slicing convention
    image_stack = image_stack[
        z_start : z_end + 1,
        y_start : y_end + 1,
        x_start : x_end + 1,
    ]
    return image_stack


def bbox_range(
    image_stack: np.ndarray, min_threshold: float
) -> Tuple[int, int, int, int, int, int]:
    """
    Calculate the bounding box range in all dimensions.

    Parameters
    ----------
    image_stack : np.ndarray
        The stack of images as a data cube.
    min_threshold : float
        The minimum threshold value for calculating the bounding box (values greater than this threshold will be retained).


    Returns
    -------
    tuple[int, int, int, int, int, int]
        The bounding box range in all dimensions.

    Examples
    --------
    >>> image_stack = np.ones((100, 100, 100))
    >>> bbox_range(image_stack, 0.5)
    (0, 99, 0, 99, 0, 99)
    """

    image_stack = image_stack > min_threshold
    # Determine the bbox range, see link
    # https://stackoverflow.com/questions/31400769/bounding-box-of-numpy-array

    # TODO: check this function with RGB image stack as input
    # N = image_stack.ndim
    N = 3
    bbox = []
    for ax in itertools.combinations(reversed(range(N)), N - 1):
        nonzero = np.any(image_stack, axis=ax)
        bbox.extend(np.where(nonzero)[0][[0, -1]])
    (z_start, z_end, y_start, y_end, x_start, x_end) = tuple(bbox)

    return (z_start, z_end, y_start, y_end, x_start, x_end)


def downscale(path_file_input: str) -> bool:
    """
    Downscale the image stack based on the provided input file.

    This function reads the input file, processes the image stack, and saves the downscaled stack.

    Parameters
    ----------
    path_file_input : str
        The path to the input file.

    Returns
    -------
    bool
        True if the processing is successful, False otherwise.
    """

    processed = False

    print(f"Processing file: {path_file_input}")

    db = ut.yaml_to_dict(Path(path_file_input))

    output_stack_type = db["output_stack_type"]
    if not (
        (output_stack_type == "downscaled")
        or (output_stack_type == "bounding_box")
        or (output_stack_type == "padded")
    ):
        raise ValueError(
            f'Invalid "output_stack_type" of "{output_stack_type}" input. Valid options inclue "downscaled", "bounding_box", or "padded".'
        )

    segmented_stack = ut.read_images(
        Path(db["image_dir"]).expanduser(), db["image_type"]
    )
    # zyx dimension ordering
    print(f"Original array size: {segmented_stack.shape}")

    dim_list = ["dz", "dy", "dx"]

    z_pad, y_pad, x_pad = [
        pad_amount(
            segmented_stack.shape[dim],
            db["resolution_output"][dim_list[dim]],
            db["resolution_input"][dim_list[dim]],
            db["downscale_tolerance"],
            db["image_limit_factor"],
        )
        for dim in range(0, 3)  # TODO should we restrict the dimensions to 3?
    ]

    padded_stack = np.pad(segmented_stack, (z_pad, y_pad, x_pad))
    print(f"New array size: {padded_stack.shape}")

    # Use list comprehension
    z_ratio, y_ratio, x_ratio = [
        db["resolution_input"][dim_list[dim]] / db["resolution_output"][dim_list[dim]]
        for dim in range(0, 3)  # TODO should we restrict the dimensions to 3
    ]

    downscaled_stack = ndimage.zoom(
        padded_stack,
        (z_ratio, y_ratio, x_ratio),
        order=0,
        mode="grid-constant",
        grid_mode=True,
    )

    print(f"Downscaled array size: {downscaled_stack.shape}")

    output_stack_type = db["output_stack_type"]
    if output_stack_type == "downscaled":
        output_stack = downscaled_stack

    else:
        cropped_stack = apply_bbox(downscaled_stack, 0)

        if output_stack_type == "bounding_box":
            output_stack = cropped_stack

        else:  # output_stack_type == "padded"
            # Pad after cropping
            (box_pad_z, box_pad_y, box_pad_x) = (
                db["padding"]["nz"],
                db["padding"]["ny"],
                db["padding"]["nx"],
            )
            padded_stack = np.pad(
                cropped_stack, ((box_pad_z,), (box_pad_y,), (box_pad_x,))
            )
            output_stack = padded_stack

    # output new downscaled stack
    out_dir = db["out_dir"]
    folder_suffix = f"{int(db['resolution_output']['dx'])}_dx"
    save_downscale_stack(output_stack, Path(out_dir), folder_suffix)

    # TODO: functionalize and test
    if db["save_npy"]:
        # npy_path = Path(out_dir).joinpath(f"{folder_suffix}.npy")
        npy_path = Path(out_dir).expanduser().joinpath(f"{folder_suffix}.npy")
        np.save(str(npy_path), output_stack)
        print(".npy file saved.")

    # TODO: functionalize and test
    if db["writeVTR"]:
        data = output_stack
        nx, ny, nz = data.shape[0], data.shape[1], data.shape[2]

        x = np.arange(0, nx + 1)
        y = np.arange(0, ny + 1)
        z = np.arange(0, nz + 1)

        # vtk_path = Path(out_dir).joinpath(f"{folder_suffix}")
        vtk_path = Path(out_dir).expanduser().joinpath(f"{folder_suffix}")
        gridToVTK(str(vtk_path), x, y, z, cellData={"imagedata": data})
        print(".vtr file saved.")

    processed = True  # overwrite, if we reach this point, all code is successful

    print(f"Finished processing file: {path_file_input}")

    return processed


def pad_amount(
    img_stack_dim_size: int,
    target_res: float,
    original_res: float,
    tolerance: float,
    limit_factor: float,
) -> tuple[int, int]:
    """
    Determine the amount of padding to add along a dimension.

    If the amount of padding is odd, the larger value will be padded to the front of the dimension.

    Parameters
    ----------
    img_stack_dim_size : int
        The size of the image stack along a dimension.
    target_res : float
        The target resolution.
    original_res : float
        The original resolution.
    tolerance : float
        The tolerance for the padding calculation.
    limit_factor : float
        The limit factor for the padding calculation.

    Returns
    -------
    tuple[int, int]
        The amount of padding to add to the front and back of the dimension.

    Examples
    --------
    >>> ds.pad_amount(100,0.62,1.0,0.01,2.0)
    (12, 12)
    """
    the_padded_size = padded_size(
        img_stack_dim_size,
        target_res,
        original_res,
        tolerance,
        limit_factor,
    )
    total_pad = the_padded_size - img_stack_dim_size
    pad = (math.ceil(total_pad / 2), math.floor(total_pad / 2))

    return pad


def padded_size(
    img_stack_dim_size: int,
    target_res: float,
    original_res: float,
    tolerance: float,
    limit_factor: float,
) -> int:
    """
    Determine the expected pad size along a dimension.

    Parameters
    ----------
    img_stack_dim_size : int
        The size of the image stack along a dimension.
    target_res : float
        The target resolution.
    original_res : float
        The original resolution.
    tolerance : float
        The tolerance for the padding calculation.
    limit_factor : float
        The limit factor for the padding calculation.

    Returns
    -------
    int
        The new dimension size after padding.

    Examples
    --------
    >>> padded_size(100, 0.5, 1.0, 0.01, 2.0)
    100
    """

    downscale_factor = float(target_res) / float(original_res)

    # start at the prior dimension size to enter the while loop
    new_dim = img_stack_dim_size - 1

    _too_high, _too_low = True, True

    _b = 0.0  # what we are comparing the result to

    while _too_high and _too_low:
        if new_dim > (img_stack_dim_size * int(limit_factor)):
            raise ValueError(
                f'Could not find even padding within "{tolerance}" for a {limit_factor}x image size. Increase tolerance or choose a different "target_res"'
            )
        new_dim += 1

        # to avoid the modulo (new_dim % downscale_factor) float error, so we must do manually
        _a_high = new_dim - (math.ceil(new_dim / downscale_factor) * downscale_factor)
        _a_low = new_dim - (math.floor(new_dim / downscale_factor) * downscale_factor)

        _too_high = not math.isclose(
            _a_high,
            _b,
            abs_tol=tolerance,
        )
        _too_low = not math.isclose(
            _a_low,
            _b,
            abs_tol=tolerance,
        )

    return new_dim


def save_downscale_stack(
    image_stack: np.ndarray, path: Path, folder_suffix: str
) -> bool:
    """
    Save the new stack as a tiff image stack.

    Parameters
    ----------
    image_stack : np.ndarray
        The stack of images as a data cube.
    path : Path
        The save path.
    folder_suffix : str
        The resolution in the suffix of the save folder "images_at_resolution_{RES_dx}".

    Returns
    -------
    bool
        True for success, False otherwise.

    Examples
    --------
    >>> image_stack = np.random.rand(100, 100, 100)
    >>> save_downscale_stack(image_stack, Path("/path/to/save"), "0.5_dx")
    Saving cropped image stack in /path/to/save > images_at_resolution_0.5_dx
    True
    """

    image_folder_name = f"images_at_resolution_{folder_suffix}"
    print(f"Saving cropped image stack in {path} > {image_folder_name}")
    ut.ndarray_to_img(
        data=image_stack,
        slice_axis=CartesianAxis3D.Z,
        parent_dir=path,
        folder_name=image_folder_name,
    )

    return True


def main():
    """
    Runs the module from the command line, invoked from pyproject.toml with 'downscale' command.

    This function sets up the command line argument parser, parses the input arguments,
    and calls the `downscale` function with the provided input file.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Examples
    --------
    To run this module from the command line with the 'downscale' command:

        $ downscale input_file.yml
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="the .yml user input file")
    args = parser.parse_args()
    input_file = args.input_file

    downscale(path_file_input=input_file)


if __name__ == "__main__":
    main()



"""Converts a semantic segmentation into a finite element mesh.
"""

import argparse

# from importlib.metadata import version
from pathlib import Path
from typing import NamedTuple

import numpy as np
import yaml

# import automesh as am
from automesh import Voxels


class AutomeshRecipe(NamedTuple):
    """
    Defines the receipe to run automesh directly from its Python API.

    Attributes
    ----------
    npy_input : Path
        The path to the numpy input file.
    output_file: Path
        The path to the mesh output file.
    remove: List[int]
        Voxel IDs to remove from the mesh [default: [0,]].
    scale_x : float
        The scaling factor along the x-axis [default: 1.0].
    scale_y : float
        The scaling factor along the y-axis [default: 1.0].
    scale_z : float
        The scaling factor along the z-axis [default: 1.0].
    translate_x : float
        The translation along the x-axis [default: 0.0].
    translate_y : float
        The translation along the y-axis [default: 0.0].
    translate_z : float
        The translation along the z-axis [default: 0.0].
    """

    npy_input: Path
    output_file: Path
    remove: list[int]
    scale_x: float = 1.0
    scale_y: float = 1.0
    scale_z: float = 1.0
    translate_x: float = 0.0
    translate_y: float = 0.0
    translate_z: float = 0.0


def validate_recipe(*, recipe: AutomeshRecipe) -> bool:
    """
    Validate the given recipe.

    Ensures that all values in the Recipe NamedTuple are valid.

    Parameters
    ----------
    recipe : AutomeshRecipe
        The Recipe NamedTuple populated with items from the user input
        .yml file.

    Returns
    -------
    bool
        True if the recipe is valid, False otherwise.

    Raises
    ------
    AssertionError
        If any of the validation checks fail.

    Examples
    --------
    >>> recipe = AutomeshRecipe(
    ...     npy_input=Path("/path/to/input.npy"),
    ...     output_file=Path("/path/to/output.inp"),
    ...     remove=[0,],
    ...     scale_x=1.0,
    ...     scale_y=1.0,
    ...     scale_z=1.0,
    ...     translate_x=0.0,
    ...     translate_y=0.0,
    ...     translate_z=0.0,
    ... )
    >>> validate_recipe(recipe=recipe)
    True
    """

    amr = recipe

    # Assure the .npy input file can be found
    assert amr.npy_input.is_file(), f"Cannot find {amr.npy_input}"

    # Assure the folder that will contain the output file can be found
    output_path = amr.output_file.parent
    assert output_path.is_dir(), f"Cannot find output path {output_path}"

    removes = amr.remove
    assert isinstance(removes, list), f"{removes} must be a list"
    all_ints = [isinstance(x, int) for x in removes]
    all_nonneg = [x >= 0 for x in removes]
    assert all(all_ints), f"{removes} must be a list of integers"
    assert all(
        all_nonneg
    ), f"{removes} must be a list of non-negative integers"

    assert amr.scale_x > 0.0, f"{amr.scale_x} must be > 0.0"
    assert amr.scale_y > 0.0, f"{amr.scale_y} must be > 0.0"
    assert amr.scale_z > 0.0, f"{amr.scale_z} must be > 0.0"

    return True


def voxel_to_mesh(*, yml_input_file: Path) -> bool:
    """
    Convert a .npy file specified in a yml recipe to a finite element mesh.

    Parameters
    ----------
    yml_input_file : Path
        The .yml recipe that specifies path variables.

    Returns
    -------
    True if successful, False otherwise.

    Raises
    ------
    FileNotFoundError
        If the input .yml file is not found.
    TypeError
        If the input file type is not supported.
    OSError
        If there is an error with the yml module.

    Examples
    --------
    >>> yml_input_file = Path("/path/to/recipe.yml")
    >>> voxel_to_mesh(yml_input_file=yml_input_file)
    0
    """

    print(f"This is {Path(__file__).resolve()}")

    fin = yml_input_file.resolve().expanduser()

    print(f"Processing file: {fin}")

    if not fin.is_file():
        raise FileNotFoundError(f"File not found: {str(fin)}")

    file_type = fin.suffix.casefold()
    supported_types = (".yaml", ".yml")

    if file_type not in supported_types:
        raise TypeError("Only file types .yaml, and .yml are supported.")

    db = []

    try:
        with open(file=fin, mode="r", encoding="utf-8") as stream:
            db = yaml.load(stream, Loader=yaml.SafeLoader)  # overwrite
    except yaml.YAMLError as error:
        print(f"Error with yml module: {error}")
        print(f"Could not open or decode: {fin}")
        raise OSError from error

    print(f"Success: database created from file: {fin}")
    print(db)

    recipe = AutomeshRecipe(
        npy_input=Path(db["npy_input"]).expanduser(),
        output_file=Path(db["output_file"]).expanduser(),
        remove=db["remove"],
        scale_x=db["scale_x"],
        scale_y=db["scale_y"],
        scale_z=db["scale_z"],
        translate_x=db["translate_x"],
        translate_y=db["translate_y"],
        translate_z=db["translate_z"],
    )

    validate_recipe(recipe=recipe)

    amr = recipe

    # Run Sculpt
    # cc = [str(recipe.sculpt_binary), "-i", str(path_sculpt_i)]
    # result = subprocess.run(cc)

    # run automesh as a subprocess
    cc = "mesh "
    cc += f"-i {amr.npy_input} "
    cc += f"-o {amr.output_file} "
    for ri in amr.remove:
        cc += f"-r {ri} "
    cc += f"--xscale {amr.scale_x} "
    cc += f"--yscale {amr.scale_y} "
    cc += f"--zscale {amr.scale_z} "
    cc += f"--xtranslate {amr.translate_x} "
    cc += f"--ytranslate {amr.translate_y} "
    cc += f"--ztranslate {amr.translate_z}"

    print("Running automesh with the following command:")
    print(f"{cc}")

    # The automesh library stores .npy arrays as [x[y[z]]], whereas the
    # image processing convention we wish to use is [z[y[x]]].  So, save
    # temporary .npy file that is reordered, and feel that to automesh
    # instead of the original .npy file.
    aa = np.load(str(amr.npy_input))
    bb = amr.npy_input.stem + "_xyz.npy"
    # temp_npy = amr.npy_input.parent.joinpath("temp.npy"
    temp_npy = amr.npy_input.parent.joinpath(bb)
    print(f"Created temporary file in xyz order for automesh: {temp_npy}")
    cc = aa.transpose(2, 1, 0)
    # TODO: Check if the z-order needs to be reveresed based on the CT serial
    # section data.
    np.save(temp_npy, cc)

    # voxels = am.Voxels.from_npy
    # voxels = Voxels.from_npy(str(amr.npy_input))
    voxels = Voxels.from_npy(str(temp_npy))
    elements = voxels.as_finite_elements(
        remove=amr.remove,
        scale=[amr.scale_x, amr.scale_y, amr.scale_z],
        translate=[amr.translate_x, amr.translate_y, amr.translate_z],
    )
    elements.write_exo(str(amr.output_file))
    print(f"Wrote output file: {amr.output_file}")

    # clean up the temporary file:
    try:
        # Delete the file
        temp_npy.unlink()
        print(f"Temporary file successfully deleted: {temp_npy}")
    except FileNotFoundError:
        print(f"Temporary file does not exist: {temp_npy}")
    except PermissionError:
        print(f"Permission denied: Unable to delete tempoary file: {temp_npy}")
    except Exception as e:
        print(f"An error occurred: {e}")

    return True  # success


def main():
    """
    Runs the module from the command line.

    This function serves as the entry point for terminal-based access to the
    module.  It processes a YAML input file to a finite element mesh.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Examples
    --------
    To run the module, use the following command in the terminal:
    $ npy_to_mesh path/to/input.yml
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="the .yml npy to mesh recipe")
    args = parser.parse_args()
    input_file = args.input_file
    input_file = Path(input_file).expanduser()

    voxel_to_mesh(yml_input_file=input_file)


if __name__ == "__main__":
    main()



"""
This module stores constant values in a single location for recon3d.

This module defines various constants and conversion factors that are used 
throughout the recon3d application. It provides a centralized location for 
these constants, making the code more maintainable and easier to understand.

Classes
-------
Constants
    Creates all constants used in this module.

Conversions
    Unit conversions.

SkimageConnectivity
    Connectivity level to be used with skimage.measure.label function in 3D.

CC3DConnectivity
    Connectivity level to be used with cc3d.connected_components function in 3D.

Attributes
----------
MODULE_NAME : str
    The name of the module.
"""

from typing import NamedTuple, Final
import numpy as np

MODULE_NAME: Final[str] = "recon3d"


class Constants(NamedTuple):
    """
    Creates all constants used in this module.

    This class defines a set of constants that are used throughout the module.
    It utilizes the `NamedTuple` from the `typing` module to create immutable
    constants.

    Attributes
    ----------
    YML_SCHEMA_VERSION : float
        The version of the YAML schema.
    module_short_name : str
        The short name of the module.
    module_prompt : str
        The prompt string for the module.
    KNUD_THOMSEN_FACTOR : float
        The Knud Thomsen factor for the surface area of a scalene ellipsoid.
    """

    YML_SCHEMA_VERSION: float = 1.0
    module_short_name: str = MODULE_NAME
    module_prompt: str = MODULE_NAME + ">"
    KNUD_THOMSEN_FACTOR: float = 1.6075  # for surface area of scalene ellipsoid

    # def __init__(self) -> None:
    #     self.YML_SCHEMA_VERSION = 1.0  # Good


class Conversions(NamedTuple):
    """
    Unit conversions.

    This class defines constants for converting between radians and degrees.
    It utilizes the `NamedTuple` from the `typing` module to create immutable
    constants.

    Attributes
    ----------
    RAD_TO_DEG : float
        Conversion factor from radians to degrees.
    DEG_TO_RAD : float
        Conversion factor from degrees to radians.
    """

    RAD_TO_DEG: float = 180 / np.pi
    DEG_TO_RAD: float = 1 / RAD_TO_DEG


class SkimageConnectivity(NamedTuple):
    """
    Connectivity level to be used with skimage.measure.label function in 3D.

    This class defines constants for different levels of connectivity that can
    be used with the `skimage.measure.label` function in 3D image processing.
    It utilizes the `NamedTuple` from the `typing` module to create immutable
    constants.

    Attributes
    ----------
    FACE_NEIGHBORS : int
        Connectivity level with 6 face neighbors.
    FACE_EDGE_NEIGHBORS : int
        Connectivity level with 6 face neighbors and 12 edge neighbors.
    FACE_EDGE_VERTEX_NEIGHBORS : int
        Connectivity level with 6 face neighbors, 12 edge neighbors, and 8 vertex neighbors.
    """

    FACE_NEIGHBORS = 1
    FACE_EDGE_NEIGHBORS = 2
    FACE_EDGE_VERTEX_NEIGHBORS = 3


class CC3DConnectivity(NamedTuple):
    """
    Connectivity level to be used with cc3d.connect_components function in 3D.

    This class defines constants for different levels of connectivity that can
    be used with the `cc3d.connected_components` function.
    It utilizes the `NamedTuple` from the `typing` module to create immutable
    constants.

    Attributes
    ----------
    FACE_NEIGHBORS : int
        Connectivity level with 6 face neighbors.
    FACE_EDGE_NEIGHBORS : int
        Connectivity level with 6 face neighbors and 12 edge neighbors.
    FACE_EDGE_VERTEX_NEIGHBORS : int
        Connectivity level with 6 face neighbors, 12 edge neighbors, and 8 vertex neighbors.
    """

    FACE_NEIGHBORS = 6
    FACE_EDGE_NEIGHBORS = 18
    FACE_EDGE_VERTEX_NEIGHBORS = 26



"""
Basic workflow for processing grayscale image stacks to labeled images for feature property measurement.

This module supports multiple phases (>2 phases) and includes simple segmentation for demonstration purposes. 
It writes outputs as image directories, but can be integrated with HDF for internal dataset storage.
"""

# TODO Integrate with HDF writing
# TODO Integrate with feature measurements
# TODO Use constants for Z axis in other module
# TODO Simplify/Impose structure for image directory (would be dataset directory in HDF)
# to remove magic numbers and indexing of phase_paths and label_paths
# TODO Devise better solution for multiphase support without if/else blocks


import argparse
import glob
from typing import Final, NamedTuple, Tuple, Type, List
import math

import h5py
import numpy as np
import numpy.typing as npt
from pathlib import Path
from PIL import Image
import skimage.io as skio
from sklearn.neighbors import NearestNeighbors as sklearnNN

from scipy import ndimage
from skimage.measure import label, regionprops
import cc3d
import fastremap

import recon3d.constants as cs
import recon3d.types
import recon3d.utility as ut
import recon3d.hdf_io as hio
from recon3d.types import *


# # ----------------------------------
# # USER INPUT DATA HARD CODES - begin
# # ----------------------------------
# # NEED TO PUT INTO A .yml file instead of hard codes
# # input_dir: str = "/Users/apolon/Desktop/test_recon/input/"
# # save_dir: str = "/Users/apolon/Desktop/test_recon/output/"

# input_dir: Final[str] = "/Users/chovey/temp/test_recon/input/"
# save_dir: Final[str] = "/Users/chovey/temp/test_recon/output/"

# file_type: Final[str] = ".tif"
# gray_path: Final[str] = f"{save_dir}/gray/"
# threshold: Final[int] = 185
# segmented_path: Final[str] = f"{save_dir}/segmented/"
# # first phase is always assumed to be the sample (1 or 255 in segmented images)
# phase_paths: np.ndarray = np.array(
#     [f"{save_dir}/sample_mask/", f"{save_dir}/void_mask/"]
# )
# # first label path is always assumed to the first phase after the sample,
# # same ordering as phase_paths
# label_paths: np.ndarray = np.array([f"{save_dir}/void_ids/"])

# # --------------------------------
# # USER INPUT DATA HARD CODES - end
# # --------------------------------


def calc_moment(
    indices: np.ndarray[np.int32],
    p: int,
    q: int,
    r: int,
    centroid_px: Centroid,
) -> float:
    """
    Calculate 3D moment invariants of a feature based on its voxel locations.
    Mask is given in index locations, so math here is all done in pixel units.

    Parameters
    ----------
    indices : np.ndarray[np.int32]
        Array of [Z,Y,X] indices corresponding to the feature's voxel locations.
        The shape of the array should be (n, 3), where n is the number of voxels of the feature.
        Example: np.array([[Z1, Y1, X1], [Z2, Y2, X2], ..., [Zn, Yn, Xn]]).
    p : int
        Order of the moment for the X dimension.
    q : int
        Order of the moment for the Y dimension.
    r : int
        Order of the moment for the Z dimension.
    centroid_px : Centroid
        Centroid location for X, Y, and Z axis, respectively, in pixel units.

    Returns
    -------
    float
        The 3D central moment (translation invariant) in pixel units normalized by object volume.

    Raises
    ------
    ValueError
        If the units of the centroid are not in pixel/voxel units.

    Examples
    --------
    >>> indices = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]], dtype=np.int32)
    >>> centroid_px = Centroid(
    ...     cx=Length(1.0, Units.VOXEL),
    ...     cy=Length(1.0, Units.VOXEL),
    ...     cz=Length(1.0, Units.VOXEL)
    ... )
    >>> calc_moment(indices, 2, 2, 2, centroid_px)
    0.66
    """

    unit = centroid_px.cx.unit
    if unit != Units.VOXEL:
        raise ValueError(
            f"Moment invariants are calculated using pixel/voxel units, but {unit} units were given"
        )

    mu = 0
    for i in range(0, indices.shape[0]):
        mu += (
            (indices[i][2] - centroid_px.cx.value)
            ** p  # x-location of voxel w.r.t x-centroid
            * (indices[i][1] - centroid_px.cy.value)
            ** q  # y-location of voxel w.r.t y-centroid
            * (indices[i][0] - centroid_px.cz.value)
            ** r  # z-location of voxel w.r.t z-centroid
        )
    return mu / len(indices)


def center_of_mass(
    indices: np.ndarray[np.int32], resolution: Resolution, origin: Origin
) -> Centroid:
    """
    Calculates the center of mass (centroid) of an instance object.
    Requires that the instance object has uniform density (is homogeneous).

    Parameters
    ----------
    indices : np.ndarray[np.int32]
        Array of [Z,Y,X] indices corresponding to the feature's voxel locations.
        The shape of the array should be (n, 3), where n is the number of voxels of the feature.
        Example: np.array([[Z1, Y1, X1], [Z2, Y2, X2], ..., [Zn, Yn, Xn]]).
    resolution : Resolution
        Resolution of the voxel grid in each dimension (dx, dy, dz).
    origin : Origin
        Origin of the voxel grid in each dimension (x0, y0, z0).

    Returns
    -------
    Centroid
        The centroid location for X, Y, and Z axis, respectively, in microns.

    Examples
    --------
    >>> indices = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]], dtype=np.int32)
    >>> resolution = Resolution(dx=Length(1.0, Units.MICRON), dy=Length(1.0, Units.MICRON), dz=Length(1.0, Units.MICRON))
    >>> origin = Origin(x0=Length(0.0, Units.MICRON), y0=Length(0.0, Units.MICRON), z0=Length(0.0, Units.MICRON))
    >>> center_of_mass(indices, resolution, origin)
    Centroid(cx=Length(value=1.0, unit=Units.MICRON), cy=Length(value=1.0, unit=Units.MICRON), cz=Length(value=1.0, unit=Units.MICRON))
    """

    centroid_zyx = np.add(
        np.multiply(
            np.mean(indices, axis=0),
            np.array(
                [
                    resolution.dz.value,
                    resolution.dy.value,
                    resolution.dx.value,
                ]
            ),
        ),
        np.array(
            [
                origin.z0.value,
                origin.y0.value,
                origin.x0.value,
            ]
        ),
    )
    cx = centroid_zyx[2]
    cy = centroid_zyx[1]
    cz = centroid_zyx[0]
    unit = resolution.dx.unit
    return Centroid(
        cx=Length(cx, unit=unit),
        cy=Length(cy, unit=unit),
        cz=Length(cz, unit=unit),
    )


def centroid_pix(
    centroid: Centroid, resolution: Resolution, origin: Origin
) -> Centroid:
    """
    Convert between real space units and pixel units.

    Parameters
    ----------
    centroid : Centroid
        The centroid location in real space units.
    resolution : Resolution
        Resolution of the voxel grid in each dimension (dx, dy, dz).
    origin : Origin
        Origin of the voxel grid in each dimension (x0, y0, z0).

    Returns
    -------
    Centroid
        The centroid location in pixel units.

    Examples
    --------
    >>> centroid = Centroid(cx=Length(10.0, Units.MICRON), cy=Length(20.0, Units.MICRON), cz=Length(30.0, Units.MICRON))
    >>> resolution = Resolution(dx=Length(1.0, Units.MICRON), dy=Length(1.0, Units.MICRON), dz=Length(1.0, Units.MICRON))
    >>> origin = Origin(x0=Length(0.0, Units.MICRON), y0=Length(0.0, Units.MICRON), z0=Length(0.0, Units.MICRON))
    >>> centroid_pix(centroid, resolution, origin)
    Centroid(cx=Length(value=10.0, unit=Units.VOXEL), cy=Length(value=20.0, unit=Units.VOXEL), cz=Length(value=30.0, unit=Units.VOXEL))
    """

    # convert centroids to normalized pixel units
    cx = (centroid.cx.value - origin.x0.value) / resolution.dx.value
    cy = (centroid.cy.value - origin.y0.value) / resolution.dy.value
    cz = (centroid.cz.value - origin.z0.value) / resolution.dz.value

    unit = Units.VOXEL

    return Centroid(
        cx=Length(cx, unit=unit),
        cy=Length(cy, unit=unit),
        cz=Length(cz, unit=unit),
    )


def ellipsoid_surface_area(ellipsoid: BestFitEllipsoid) -> Area:
    """
    Calculate the surface area of the best fit ellipsoid using the Knud Thomsen formula.
    Applicable to scalene ellipsoids (semi-axes a > b > c), with a relative error of at most 1.061%.

    Parameters
    ----------
    ellipsoid : BestFitEllipsoid
        The best fit ellipsoid with semi-axes lengths a, b, and c.

    Returns
    -------
    Area
        The surface area of the ellipsoid in microns squared.

    Examples
    --------
    >>> ellipsoid = BestFitEllipsoid(
    ...     a=EllipsoidAxis(length=Length(5.0, Units.MICRON), orientation=UnitVector(u=1, v=0, w=0)),
    ...     b=EllipsoidAxis(length=Length(3.0, Units.MICRON), orientation=UnitVector(u=0, v=1, w=0)),
    ...     c=EllipsoidAxis(length=Length(2.0, Units.MICRON), orientation=UnitVector(u=0, v=0, w=1))
    ... )
    >>> ellipsoid_surface_area(ellipsoid)
    Area(value=134.8149867941625, unit_squared=<Units.MICRON: 'micron'>)
    """
    unit = ellipsoid.a.length.unit
    p = cs.Constants().KNUD_THOMSEN_FACTOR
    a, b, c = (
        ellipsoid.a.length.value,
        ellipsoid.b.length.value,
        ellipsoid.c.length.value,
    )

    surface_area = (
        4 * math.pi * ((((a**p * b**p) + (a**p * c**p) + (b**p * c**p)) / 3) ** (1 / p))
    )
    return Area(value=surface_area, unit_squared=unit)


def ellipsoid_volume(ellipsoid: BestFitEllipsoid) -> Volume:
    """
    Calculate the volume of the best fit ellipsoid using semi-axes lengths.

    Parameters
    ----------
    ellipsoid : BestFitEllipsoid
        The best fit ellipsoid with semi-axes lengths a, b, and c.

    Returns
    -------
    Volume
        The volume of the ellipsoid in microns cubed.

    Examples
    --------
    >>> ellipsoid = BestFitEllipsoid(
    ...     a=EllipsoidAxis(length=Length(5.0, Units.MICRON), orientation=UnitVector(u=1, v=0, w=0)),
    ...     b=EllipsoidAxis(length=Length(3.0, Units.MICRON), orientation=UnitVector(u=0, v=1, w=0)),
    ...     c=EllipsoidAxis(length=Length(2.0, Units.MICRON), orientation=UnitVector(u=0, v=0, w=1))
    ... )
    >>> ellipsoid_volume(ellipsoid)
    Volume(value=125.66370614359171, unit_cubed=<Units.MICRON: 'micron'>)
    """

    unit = ellipsoid.a.length.unit
    a, b, c = (
        ellipsoid.a.length.value,
        ellipsoid.b.length.value,
        ellipsoid.c.length.value,
    )

    volume = 4 / 3 * math.pi * a * b * c

    return Volume(value=volume, unit_cubed=unit)


def equivalent_spherical_diameter(
    n_voxels: int,
    resolution: Resolution,
) -> Length:
    """
    Calculate the equivalent spherical diameter of a feature.

    Parameters
    ----------
    n_voxels : int
        The number of voxels in the feature.
    resolution : Resolution
        The resolution of the voxel grid in each dimension (dx, dy, dz).

    Returns
    -------
    Length
        The equivalent spherical diameter of the feature in the same units as the resolution.

    Examples
    --------
    >>> resolution = Resolution(
    ...     dx=Length(1.0, Units.MICRON),
    ...     dy=Length(1.0, Units.MICRON),
    ...     dz=Length(1.0, Units.MICRON)
    ... )
    >>> equivalent_spherical_diameter(1000, resolution)
    Length(value=12.407009817988, unit=<Units.MICRON: 'micron'>)
    """

    value = 2 * np.power(
        (
            (3 / (4 * np.pi))
            * (
                n_voxels
                * resolution.dx.value
                * resolution.dy.value
                * resolution.dz.value
            )
        ),
        (1 / 3),
    )
    return Length(value=value, unit=resolution.dx.unit)


def fit_ellipsoid(
    indices: np.ndarray,
    centroid: Centroid,
    resolution: Resolution,
    origin: Origin,
) -> BestFitEllipsoid:
    """
    Fit a 3D ellipsoid to a set of voxel indices.

    Parameters
    ----------
    indices : np.ndarray
        Array of [Z,Y,X] indices corresponding to the feature's voxel locations.
        The shape of the array should be (n, 3), where n is the number of voxels of the feature.
    centroid : Centroid
        Centroid location for X, Y, and Z axis, respectively, in spatial coordinates (e.g., microns).
    resolution : Resolution
        Voxel resolution for X, Y, and Z dimensions, respectively, in length units (e.g., microns).
    origin : Origin
        Optional origin in spatial coordinates if the volume has been previously transformed.

    Returns
    -------
    BestFitEllipsoid
        The best fit ellipsoid with semi-axes lengths and orientations.

    Examples
    --------
    >>> indices = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]], dtype=np.int32)
    >>> centroid = Centroid(cx=Length(1.0, Units.MICRON), cy=Length(1.0, Units.MICRON), cz=Length(1.0, Units.MICRON))
    >>> resolution = Resolution(dx=Length(1.0, Units.MICRON), dy=Length(1.0, Units.MICRON), dz=Length(1.0, Units.MICRON))
    >>> origin = Origin(x0=Length(0.0, Units.MICRON), y0=Length(0.0, Units.MICRON), z0=Length(0.0, Units.MICRON))
    >>> fit_ellipsoid(indices, centroid, resolution, origin)
    BestFitEllipsoid(a=EllipsoidAxis(length=Length(value=80.146386085398, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=-0.25175905655605574, v=-0.5305735687527386, w=-0.8093880809494217)), b=EllipsoidAxis(length=Length(value=1.2477168951003976, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=-0.8774683531474325, v=-0.227651095921704, w=0.4221661613042665)), c=EllipsoidAxis(length=Length(value=2.380014006104306e-07, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.4082482904637247, v=-0.816496580927762, w=0.4082482904639296)))
    """

    if not resolution.dx == resolution.dy == resolution.dz:
        raise NotImplementedError(
            "Only isotropic voxels are currently supported for axis length calculation"
        )

    unit = resolution.dx.unit

    # convert centroids to normalized pixel units
    cen_px = centroid_pix(centroid, resolution, origin)

    # Get moments
    s_xx = calc_moment(indices, 2, 0, 0, cen_px)
    s_yy = calc_moment(indices, 0, 2, 0, cen_px)
    s_zz = calc_moment(indices, 0, 0, 2, cen_px)
    s_xy = calc_moment(indices, 1, 1, 0, cen_px)
    s_xz = calc_moment(indices, 1, 0, 1, cen_px)
    s_yz = calc_moment(indices, 0, 1, 1, cen_px)

    # Create matrix of 3D moments
    s_ij = np.array(
        [
            [s_xx, s_xy, s_xz],
            [s_xy, s_yy, s_yz],
            [s_xz, s_yz, s_zz],
        ]
    )
    eig_val, eig_vec = np.linalg.eig(s_ij)

    semi_ax_lengths_px = np.sqrt(5 * eig_val)  # still in pixel units
    sort = np.argsort(semi_ax_lengths_px)[::-1]  # sort descending axis lengths

    # TODO: math to account for change in length as function
    #       as function of axis orientation in anisotropic voxels
    #       remove NotImplementedError above
    semi_ax_lengths = np.array(
        [
            semi_ax_lengths_px[sort[0]] * resolution.dx.value,
            semi_ax_lengths_px[sort[1]] * resolution.dx.value,
            semi_ax_lengths_px[sort[2]] * resolution.dx.value,
        ]
    )
    ## some axes lengths are NAN, should overwrite to 0.0
    semi_ax_lengths = np.nan_to_num(semi_ax_lengths, nan=0.0)

    ## from docs on np.linalg.eig: The normalized (unit “length”) eigenvectors,
    ## such that the column eigenvectors[:,i] is the eigenvector corresponding to the eigenvalue eigenvalues[i].
    ax_vectors = np.array(
        [
            eig_vec[:, sort[0]],
            eig_vec[:, sort[1]],
            eig_vec[:, sort[2]],
        ]
    )
    ax_vectors = np.reshape(ax_vectors, (9))

    a_vector = UnitVector(
        u=ax_vectors[0],
        v=ax_vectors[1],
        w=ax_vectors[2],
    )
    a = EllipsoidAxis(
        length=Length(semi_ax_lengths[0], unit=unit),
        orientation=a_vector,
    )
    b_vector = UnitVector(
        u=ax_vectors[3],
        v=ax_vectors[4],
        w=ax_vectors[5],
    )
    b = EllipsoidAxis(
        length=Length(semi_ax_lengths[1], unit=unit),
        orientation=b_vector,
    )
    c_vector = UnitVector(
        u=ax_vectors[6],
        v=ax_vectors[7],
        w=ax_vectors[8],
    )
    c = EllipsoidAxis(
        length=Length(semi_ax_lengths[2], unit=unit),
        orientation=c_vector,
    )

    ellipsoid = BestFitEllipsoid(a=a, b=b, c=c)
    return ellipsoid


def instance_indices(
    instance_stack: InstanceImageStack,
) -> InstanceIndices:
    """
    Returns an array of instances and a list of N-dimensional arrays of indices
    of the unique values in the instance stack.

    Parameters
    ----------
    instance_stack : InstanceImageStack
        Array with arbitrary dimensions.

    Returns
    -------
    InstanceIndices
        - 1D-array of sorted unique values.
        - List of arrays. Each array contains the indices where a given value in x is found.

    Examples
    --------
    >>> instance_stack = InstanceImageStack(
    ...     name="example_stack",
    ...     metadata=MetaData(
    ...         data_volume=DataVolume(x_width=2, y_height=2, z_image_count=1),
    ...         resolution=Resolution(
    ...             dx=Length(1.0, Units.MICRON),
    ...             dy=Length(1.0, Units.MICRON),
    ...             dz=Length(1.0, Units.MICRON)
    ...         ),
    ...         pixel_units=Units.MICRON,
    ...         origin=Origin(
    ...             x0=Length(0.0, Units.MICRON),
    ...             y0=Length(0.0, Units.MICRON),
    ...             z0=Length(0.0, Units.MICRON)
    ...         )
    ...     ),
    ...     data=np.array([[1, 2, 2], [3, 1, 1]]),
    ...     nlabels=3,
    ...     min_feature_size=1
    ... )
    >>> instance_indices(instance_stack)
    InstanceIndices(
        source_name='example_stack',
        indices=[array([[0, 0], [1, 1], [1, 2]]), array([[0, 1], [0, 2]]), array([[1, 0]])],
        labels=InstanceLabels(data=np.array([1, 2, 3]))
    )
    """

    x = instance_stack.data
    x_flat = x.ravel()
    ix_flat = np.argsort(x_flat)
    u, ix_u = fastremap.unique(x_flat[ix_flat], return_index=True)
    ix_ndim = np.unravel_index(ix_flat, x.shape)
    ix_ndim = np.c_[ix_ndim] if x.ndim > 1 else ix_flat
    indices = np.split(ix_ndim, ix_u[1:])
    labels = u
    instance_ids = InstanceIndices(
        source_name=instance_stack.name,
        indices=indices,
        labels=InstanceLabels(data=labels),
    )
    return instance_ids


def minimum_size_filter(
    initial_stack: InstanceImageStack,
    initial_indices: InstanceIndices,
) -> Tuple[InstanceImageStack, InstanceIndices]:
    """
    Removes features below minimum size from InstanceIndices and corresponding
    InstanceImageStack. Should be done before shape metrics are determined.

    Parameters
    ----------
    initial_stack : InstanceImageStack
        The initial instance image stack.
    initial_indices : InstanceIndices
        The initial instance indices.

    Returns
    -------
    Tuple[InstanceImageStack, InstanceIndices]
        The filtered instance image stack and the filtered instance indices.

    Raises
    ------
    ValueError
        If the names of the initial stack and indices do not match.

    Examples
    --------
    >>> initial_stack = InstanceImageStack(
    ...     name="example_stack",
    ...     metadata=MetaData(
    ...         data_volume=DataVolume(x_width=4, y_height=2, z_image_count=1),
    ...         resolution=Resolution(
    ...             dx=Length(1.0, Units.MICRON),
    ...             dy=Length(1.0, Units.MICRON),
    ...             dz=Length(1.0, Units.MICRON)
    ...         ),
    ...         pixel_units=Units.MICRON,
    ...         origin=Origin(
    ...             x0=Length(0.0, Units.MICRON),
    ...             y0=Length(0.0, Units.MICRON),
    ...             z0=Length(0.0, Units.MICRON)
    ...         )
    ...     ),
    ...     data=np.array([[1, 2, 2, 1], [3, 1, 1, 1]]),
    ...     nlabels=3,
    ...     min_feature_size=2
    ... )
    >>> initial_indices = instance_indices(initial_stack)
    >>> minimum_size_filter(initial_stack, initial_indices)
    2 connected components remaining after minimum size filter of 2 voxels
    (InstanceImageStack(name='example_stack', metadata=MetaData(data_volume=DataVolume(x_width=4, y_height=2, z_image_count=1), resolution=Resolution(dx=Length(value=1.0, unit=<Units.MICRON: 'micron'>), dy=Length(value=1.0, unit=<Units.MICRON: 'micron'>), dz=Length(value=1.0, unit=<Units.MICRON: 'micron'>)), pixel_units=<Units.MICRON: 'micron'>, origin=Origin(x0=Length(value=0.0, unit=<Units.MICRON: 'micron'>), y0=Length(value=0.0, unit=<Units.MICRON: 'micron'>), z0=Length(value=0.0, unit=<Units.MICRON: 'micron'>))), data=array([[1, 0, 0, 1],[2, 1, 1, 1]], dtype=int8), nlabels=2, min_feature_size=2), InstanceIndices(source_name='example_stack', labels=InstanceLabels(data=array([0, 1, 2], dtype=int8)), indices=[array([[0, 1],[0, 2]]), array([[0, 0],[0, 3],[1, 1],[1, 2],[1, 3]]), array([[1, 0]])]))
    """

    if not initial_stack.name == initial_indices.source_name:
        raise ValueError(
            f"""Initial InstanceImageStack name ({initial_stack.name}) and 
            InstanceIndices name ({initial_indices.source_name}) do not match.
            Function inputs must be from same dataset."""
        )

    n_voxels = num_voxels(initial_indices)
    n_voxels_vals = np.asarray([i.value for i in n_voxels])
    invalid_instance_ids = np.nonzero(n_voxels_vals < initial_stack.min_feature_size)[0]
    masked_instance_ids = fastremap.mask(initial_stack.data, invalid_instance_ids)
    remapped_instance_ids, _ = fastremap.renumber(
        masked_instance_ids,
        preserve_zero=True,
    )
    nlabels = np.max(np.unique(remapped_instance_ids))

    filtered_stack = InstanceImageStack(
        name=initial_stack.name,
        metadata=initial_stack.metadata,
        data=remapped_instance_ids,
        nlabels=nlabels,
        min_feature_size=initial_stack.min_feature_size,
    )

    filtered_indices = instance_indices(instance_stack=filtered_stack)

    print(
        f"\t\t{nlabels} connected components remaining after minimum size filter of {initial_stack.min_feature_size} voxels"
    )
    return filtered_stack, filtered_indices


# TODO
def map_features_to_voxels():
    """To come."""


def nearest_neighbor_distance(
    centroids: Centroids,
) -> NthNearestNeighbors:
    """
    Calculate the nearest neighbor distance for a set of centroids.
    With nth_nearest_neighbor=1 returning the features themselves
    (distances=0, neighbor is themselves).

    Parameters
    ----------
    centroids : Centroids
        The centroids for which to calculate the nearest neighbor distance.

    Returns
    -------
    NthNearestNeighbors
        The nth nearest neighbors, distances, and instance IDs.

    Examples
    --------
    >>> centroids = Centroids(data=[
    ...     Centroid(cx=Length(0.0, Units.MICRON), cy=Length(0.0, Units.MICRON), cz=Length(0.0, Units.MICRON)),
    ...     Centroid(cx=Length(1.0, Units.MICRON), cy=Length(1.0, Units.MICRON), cz=Length(1.0, Units.MICRON)),
    ...     Centroid(cx=Length(2.0, Units.MICRON), cy=Length(2.0, Units.MICRON), cz=Length(2.0, Units.MICRON))
    ... ])
    >>> nearest_neighbor_distance(centroids)
    NthNearestNeighbors(nth_nearest=2, distances=[Length(value=1.7320508075688772, unit=<Units.MICRON: 'micron'>), Length(value=1.7320508075688772, unit=<Units.MICRON: 'micron'>), Length(value=1.7320508075688772, unit=<Units.MICRON: 'micron'>)], instance_id=array([0, 0, 1]))
    """

    neighbor = sklearnNN(n_neighbors=2)
    centroid_array = ut.centroids_to_ndarray(centroids=centroids)
    neighbor.fit(centroid_array)
    neighbor_dist, neighbor_index = neighbor.kneighbors(
        centroid_array,
        return_distance=True,
    )
    distances = neighbor_dist[:, 1]
    instance_id = neighbor_index[:, 1]
    distances[0], instance_id[0] = 0, 0  # set feature 0 to 0 distance and neighbor 0

    unit = centroids.data[0].cx.unit
    distances_list = [Length(value=i, unit=unit) for i in distances.tolist()]
    nth_nearest = NthNearestNeighbors(
        nth_nearest=2,
        distances=distances_list,
        instance_id=instance_id,
    )
    return nth_nearest


def num_voxels(indices: InstanceIndices) -> list[NVoxel]:
    """
    Returns the number of voxels for each instance ID from an InstanceImageStack.

    Parameters
    ----------
    indices : InstanceIndices
        The instance indices containing the indices of each instance.

    Returns
    -------
    List[NVoxel]
        A list of NVoxel objects, each representing the number of voxels for an instance.

    Examples
    --------
    >>> indices = InstanceIndices(
    ...     source_name="example_stack",
    ...     indices=[np.array([[0, 0], [1, 1]]), np.array([[0, 1], [1, 2]])],
    ...     labels=InstanceLabels(data=np.array([1, 2]))
    ... )
    >>> num_voxels(indices)
    [NVoxel(value=2, unit=<Units.VOXEL: 'voxel'>), NVoxel(value=2, unit=<Units.VOXEL: 'voxel'>)]
    """

    indices = indices.indices
    data = [NVoxel(value=len(i)) for i in indices]
    # n_voxels = np.asarray(data, dtype=NVoxel)
    n_voxels = data
    return n_voxels


def process_image_stack(yml_input_file: Path) -> SemanticImageStack:
    """
    Given a yml input file, converts it to a SemanticImageStack.

    Parameters
    ----------
    yml_input_file : Path
        The fully pathed yml file that specifies settings.

    Returns
    -------
    SemanticImageStack
        The SemanticImageStack object.

    Examples
    --------
    >>> yml_input_file = Path("path/to/input.yml")
    >>> semantic_image_stack = process_image_stack(yml_input_file)
    """

    print(f"Processing specification file: {yml_input_file}")
    db = ut.yaml_to_dict(yml_input_file)

    file_type = db["semantic_images_type"]
    valid_extensions = (".tif", ".tiff")

    if file_type not in valid_extensions:
        raise ValueError("Error: 'image_type' must be '.tif' or '.tiff'.")

    path_input = Path(db["semantic_images_dir"]).expanduser()
    if not path_input.is_dir():
        raise ValueError(f"Error, 'image_dir' of {path_input} not found.")
    print(f"Input path: {path_input}")

    path_output = Path(db["out_dir"]).expanduser()
    if not path_output.is_dir():
        Path(path_output).mkdir(
            parents=True, exist_ok=True
        )  # makes directory if it doesn't exist
    print(f"Output path: {path_output}")

    # semantic_seg_dict = db["class_labels"]

    # # load images
    # imageStack = skio.imread_collection(tiff_files, conserve_memory=True)
    data = ut.read_images(path_input, file_type)
    semantic_seg_stack_name = db["semantic_imagestack_name"]

    # create meta data
    (nz, ny, nx) = data.shape
    data_volume = DataVolume(z_image_count=nz, y_height=ny, x_width=nx)
    print(
        f"image stack, {semantic_seg_stack_name}, has dimensions (num_images, row, col): {data.shape}"
    )

    pixel_units = db["pixel_units"]
    valid_units = set(item.value for item in Units)
    if pixel_units not in valid_units:
        raise ValueError(
            f"Error, '{pixel_units}' is not a valid unit, accepted units are: {valid_units}"
        )
    pixel_units = Units(db["pixel_units"])

    dx, dy, dz = (
        db["voxel_size"]["dx"],
        db["voxel_size"]["dy"],
        db["voxel_size"]["dz"],
    )
    resolution = Resolution(
        dx=Length(dx, unit=pixel_units),
        dy=Length(dy, unit=pixel_units),
        dz=Length(dz, unit=pixel_units),
    )

    x0, y0, z0 = (
        db["origin"]["x0"],
        db["origin"]["y0"],
        db["origin"]["z0"],
    )
    origin = Origin(
        x0=Length(x0, unit=pixel_units),
        y0=Length(y0, unit=pixel_units),
        z0=Length(z0, unit=pixel_units),
    )

    meta = MetaData(
        data_volume=data_volume,
        resolution=resolution,
        pixel_units=pixel_units,
        origin=origin,
    )

    # # Optional, grey image stack
    # # TODO: test functionality
    # if db["grey_image_dir"]:
    #     path_grey_image_dir = Path(db["path_grey_image_dir"]).expanduser()
    #     assert path_grey_image_dir.is_dir(), "Error, 'image_dir' not found."
    #     grey_image_name = "Raw greyscale"
    #     grey_data = ut.read_images(path_grey_image_dir, file_type)
    #     assert data.shape == grey_data.shape
    #     semantic_seg_image_stack = ImageStack(
    #         name=grey_image_name, metadata=meta, data=grey_data
    #     )

    # create SemanticImageStack
    semantic_seg_image_stack = SemanticImageStack(
        name=semantic_seg_stack_name, metadata=meta, data=data
    )

    return semantic_seg_image_stack  # , path_file_output)


def save_instance_images(
    input_volumes: list[InstanceImageStack], save_path: Path
) -> bool:
    """
    Save InstanceImageStack objects to image files.

    This function takes a list of InstanceImageStack objects and saves each one as a series of image files.
    The images are saved in the specified directory, with each stack saved in its own subdirectory.

    Parameters
    ----------
    input_volumes : list of InstanceImageStack
        List of InstanceImageStack objects to be saved as images.
    save_path : Path
        The directory where the images will be saved.

    Returns
    -------
    bool
        True if the images were saved successfully, False otherwise.

    Examples
    --------
    >>> input_volumes = [
    ...     InstanceImageStack(
    ...         name="example_stack_1",
    ...         metadata=MetaData(
    ...             data_volume=DataVolume(x_width=256, y_height=256, z_image_count=10),
    ...             resolution=Resolution(
    ...                 dx=Length(1.0, Units.MICRON),
    ...                 dy=Length(1.0, Units.MICRON),
    ...                 dz=Length(1.0, Units.MICRON)
    ...             ),
    ...             pixel_units=Units.MICRON,
    ...             origin=Origin(
    ...                 x0=Length(0.0, Units.MICRON),
    ...                 y0=Length(0.0, Units.MICRON),
    ...                 z0=Length(0.0, Units.MICRON)
    ...             )
    ...         ),
    ...         data=np.random.rand(10, 256, 256),
    ...         nlabels=3,
    ...         min_feature_size=2
    ...     ),
    ...     InstanceImageStack(
    ...         name="example_stack_2",
    ...         metadata=MetaData(
    ...             data_volume=DataVolume(x_width=256, y_height=256, z_image_count=10),
    ...             resolution=Resolution(
    ...                 dx=Length(1.0, Units.MICRON),
    ...                 dy=Length(1.0, Units.MICRON),
    ...                 dz=Length(1.0, Units.MICRON)
    ...             ),
    ...             pixel_units=Units.MICRON,
    ...             origin=Origin(
    ...                 x0=Length(0.0, Units.MICRON),
    ...                 y0=Length(0.0, Units.MICRON),
    ...                 z0=Length(0.0, Units.MICRON)
    ...             )
    ...         ),
    ...         data=np.random.rand(10, 256, 256),
    ...         nlabels=3,
    ...         min_feature_size=2
    ...     )
    ... ]
    >>> save_path = Path("path/to/save")
    >>> save_instance_images(input_volumes, save_path)
    True
    """

    for instance_volumes in input_volumes:
        ut.ndarray_to_img(
            data=instance_volumes.data,
            slice_axis=recon3d.types.CartesianAxis3D.Z,
            parent_dir=save_path,
            folder_name=instance_volumes.name,
            file_type=".tif",
            pad_length=4,
        )


def semantic_to_instance(
    semantic_stack: SemanticImageStack,
    instance_name: str,
    instance_value: int,
    min_feature_size: int,
) -> InstanceImageStack:
    """
    Isolate each unique class in the semantic segmentation (e.g., Cats or Dogs)
    and create the Instance Image Stacks.

    Parameters
    ----------
    semantic_stack : SemanticImageStack
        The semantic image stack containing the segmented data.
    instance_name : str
        The name for the instance image stack.
    instance_value : int
        The value in the semantic stack representing the class to isolate.
    min_feature_size : int
        The minimum feature size to retain in the instance image stack.

    Returns
    -------
    InstanceImageStack
        The instance image stack with isolated features.

    Examples
    --------
    >>> semantic_stack = SemanticImageStack(
    ...     name="example_semantic_stack",
    ...     metadata=MetaData(
    ...         data_volume=DataVolume(z_image_count=10, y_height=256, x_width=256),
    ...         resolution=Resolution(
    ...             dx=Length(1.0, Units.MICRON),
    ...             dy=Length(1.0, Units.MICRON),
    ...             dz=Length(1.0, Units.MICRON)
    ...         ),
    ...         pixel_units=Units.MICRON,
    ...         origin=Origin(
    ...             x0=Length(0.0, Units.MICRON),
    ...             y0=Length(0.0, Units.MICRON),
    ...             z0=Length(0.0, Units.MICRON)
    ...         )
    ...     ),
    ...     data=np.random.randint(0, 2, (10, 256, 256))
    ... )
    >>> instance_stack = semantic_to_instance(
    ...     semantic_stack,
    ...     instance_name="example_instance_stack",
    ...     instance_value=1,
    ...     min_feature_size=10
    ... )
    >>> print(instance_stack)
    InstanceImageStack(name='example_instance_stack', metadata=MetaData(data_volume=DataVolume(z_image_count=10, y_height=256, x_width=256), resolution=Resolution(dx=Length(value=1.0, unit=<Units.MICRON: 'micron'>), dy=Length(value=1.0, unit=<Units.MICRON: 'micron'>), dz=Length(value=1.0, unit=<Units.MICRON: 'micron'>)), pixel_units=<Units.MICRON: 'micron'>, origin=Origin(x0=Length(value=0.0, unit=<Units.MICRON: 'micron'>), y0=Length(value=0.0, unit=<Units.MICRON: 'micron'>), z0=Length(value=0.0, unit=<Units.MICRON: 'micron'>))), data=array(...), nlabels=..., min_feature_size=10)
    """

    # Create a boolean mask of the array
    masked_stack = semantic_stack.data == instance_value

    print(f"\tlabelling connected components in '{instance_name}'")

    # Use cc3d for connected components labeling
    cc3d_instance_stack, cc3d_nlabels = cc3d.connected_components(
        masked_stack,
        connectivity=cs.CC3DConnectivity.FACE_NEIGHBORS,
        return_N=True,
    )
    print(
        f"\t\twith cc3d package, found {cc3d_nlabels} connected components in '{instance_name}'"
    )
    return InstanceImageStack(
        name=instance_name,
        data=cc3d_instance_stack,
        metadata=semantic_stack.metadata,
        nlabels=cc3d_nlabels,
        min_feature_size=min_feature_size,
    )


##### MAIN FUNCTIONS


# TODO remove hardcoding here with yml config type
def instance_analysis_included(settings: dict, key: str) -> bool:
    """
    Checks YAML settings to determine if instance analysis should be performed.

    This function checks the provided settings dictionary to see if instance analysis
    is included for the specified key.

    Parameters
    ----------
    settings : dict
        The settings dictionary, typically loaded from a YAML file.
    key : str
        The key representing the class label to check for instance analysis inclusion.

    Returns
    -------
    bool
        True if instance analysis is included for the specified key, False otherwise.

    Examples
    --------
    >>> settings = {
    ...     "class_labels": {
    ...         "cat": {
    ...             "instance_analysis": {
    ...                 "include": True
    ...             }
    ...         },
    ...         "dog": {
    ...             "instance_analysis": {
    ...                 "include": False
    ...             }
    ...         }
    ...     }
    ... }
    >>> instance_analysis_included(settings, "cat")
    True
    >>> instance_analysis_included(settings, "dog")
    False
    >>> instance_analysis_included(settings, "bird")
    False
    """

    labels = settings["class_labels"]
    try:
        return labels[key]["instance_analysis"]["include"]
    except KeyError:
        return False


def instance_properties(
    instance_stack: InstanceImageStack,
    inst_indices: InstanceIndices,
) -> InstanceProperties:
    """
    Calculate various properties for each instance in an InstanceImageStack.

    Parameters
    ----------
    instance_stack : InstanceImageStack
        The instance image stack containing the segmented data.
    inst_indices : InstanceIndices
        The instance indices containing the indices of each instance.

    Returns
    -------
    InstanceProperties
        The properties of each instance, including number of voxels, equivalent spherical diameters,
        centroids, ellipsoids, surface areas, and volumes.

    Examples
    --------
    >>> instance_stack = InstanceImageStack(
    ...     name="example_instance_stack",
    ...     metadata=MetaData(
    ...         data_volume=DataVolume(z_image_count=10, y_height=256, x_width=256),
    ...         resolution=Resolution(
    ...             dx=Length(1.0, Units.MICRON),
    ...             dy=Length(1.0, Units.MICRON),
    ...             dz=Length(1.0, Units.MICRON)
    ...         ),
    ...         pixel_units=Units.MICRON,
    ...         origin=Origin(
    ...             x0=Length(0.0, Units.MICRON),
    ...             y0=Length(0.0, Units.MICRON),
    ...             z0=Length(0.0, Units.MICRON)
    ...         )
    ...     ),
    ...     data=np.random.randint(0, 2, (10, 256, 256)),
    ...     nlabels=3,
    ...     min_feature_size=2
    ... )
    >>> inst_indices = InstanceIndices(
    ...     source_name="example_instance_stack",
    ...     indices=[np.array([[0, 0], [1, 1]]), np.array([[0, 1], [1, 2]])],
    ...     labels=InstanceLabels(data=np.array([1, 2]))
    ... )
    >>> properties = instance_properties(instance_stack, inst_indices)
    >>> print(properties)
    InstanceProperties(source_name='example_instance_stack', labels=InstanceLabels(data=array([1, 2])), n_voxels=[NVoxel(value=0), NVoxel(value=2), NVoxel(value=2)], equivalent_sphere_diameters=[Length(value=0.0, unit=<Units.MICRON: 'micron'>), Length(value=1.240700981798799, unit=<Units.MICRON: 'micron'>), Length(value=1.240700981798799, unit=<Units.MICRON: 'micron'>)], centroids=Centroids(data=[Centroid(cx=Length(value=0.0, unit=<Units.MICRON: 'micron'>), cy=Length(value=0.0, unit=<Units.MICRON: 'micron'>), cz=Length(value=0.0, unit=<Units.MICRON: 'micron'>)), Centroid(cx=Length(value=0.5, unit=<Units.MICRON: 'micron'>), cy=Length(value=0.5, unit=<Units.MICRON: 'micron'>), cz=Length(value=0.5, unit=<Units.MICRON: 'micron'>)), Centroid(cx=Length(value=1.5, unit=<Units.MICRON: 'micron'>), cy=Length(value=1.5, unit=<Units.MICRON: 'micron'>), cz=Length(value=1.5, unit=<Units.MICRON: 'micron'>))]), ellipsoids=BestFitEllipsoids(data=[BestFitEllipsoid(a=EllipsoidAxis(length=Length(value=0.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.0, v=0.0, w=0.0)), b=EllipsoidAxis(length=Length(value=0.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.0, v=0.0, w=0.0)), c=EllipsoidAxis(length=Length(value=0.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.0, v=0.0, w=0.0))]), BestFitEllipsoid(a=EllipsoidAxis(length=Length(value=1.240700981798799, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=1.0, v=0.0, w=0.0)), b=EllipsoidAxis(length=Length(value=0.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.0, v=1.0, w=0.0)), c=EllipsoidAxis(length=Length(value=0.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.0, v=0.0, w=1.0))]), BestFitEllipsoid(a=EllipsoidAxis(length=Length(value=1.240700981798799, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=1.0, v=0.0, w=0.0)), b=EllipsoidAxis(length=Length(value=0.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.0, v=1.0, w=0.0)), c=EllipsoidAxis(length=Length(value=0.0, unit=<Units.MICRON: 'micron'>), orientation=UnitVector(u=0.0, v=0.0, w=1.0))])]), surface_areas=EllipsoidSurfaceAreas(data=[Area(value=0.0, unit_squared=<Units.MICRON: 'micron'>), Area(value=19.739208802178716, unit_squared=<Units.MICRON: 'micron'>), Area(value=19.739208802178716, unit_squared=<Units.MICRON: 'micron'>)]), volumes=EllipsoidVolumes(data=[Volume(value=0.0, unit_cubed=<Units.MICRON: 'micron'>), Volume(value=4.1887902047863905, unit_cubed=<Units.MICRON: 'micron'>), Volume(value=4.1887902047863905, unit_cubed=<Units.MICRON: 'micron'>)]))
    """

    num_labels = instance_stack.nlabels
    # feature 0 is the background, and we want to include the last label
    stack_n_voxels = num_voxels(indices=inst_indices)
    stack_eq_diameters = np.zeros(num_labels + 1, dtype=object)
    stack_centroids = np.zeros(num_labels + 1, dtype=object)
    stack_ellipsoids = np.zeros(num_labels + 1, dtype=object)
    stack_surface_areas = np.zeros(num_labels + 1, dtype=object)
    stack_volumes = np.zeros(num_labels + 1, dtype=object)

    resolution = instance_stack.metadata.resolution
    origin = instance_stack.metadata.origin

    # null types
    null_length = Length(value=0.0, unit=resolution.dx.unit)
    null_area = Area(value=0.0, unit_squared=resolution.dx.unit)
    null_volume = Volume(value=0.0, unit_cubed=resolution.dx.unit)
    null_axis = EllipsoidAxis(
        length=null_length,
        orientation=UnitVector(
            u=0.0,
            v=0.0,
            w=0.0,
        ),
    )
    # set first entry to correct type of zero for feature 0
    stack_n_voxels[0] = NVoxel(value=0)
    stack_eq_diameters[0] = null_length
    stack_centroids[0] = Centroid(
        cx=null_length,
        cy=null_length,
        cz=null_length,
    )
    stack_ellipsoids[0] = BestFitEllipsoid(
        a=null_axis,
        b=null_axis,
        c=null_axis,
    )
    stack_surface_areas[0] = null_area
    stack_volumes[0] = null_volume

    for instance in range(1, num_labels + 1):
        if ((instance - 1) % 1000) == 0:
            print(f"\t\tAnalyzing feature {instance} of {num_labels}")
        indices = inst_indices.indices[instance]
        n_voxels = stack_n_voxels[instance].value

        # Equivalent spherical diameter
        stack_eq_diameters[instance] = equivalent_spherical_diameter(
            n_voxels=n_voxels,
            resolution=resolution,
        )

        # Centroid (real-space coordinates)
        centroid = center_of_mass(
            indices=indices,
            resolution=resolution,
            origin=origin,
        )
        stack_centroids[instance] = centroid

        # Ellipsoid properties
        ellipsoid = fit_ellipsoid(
            indices=indices,
            centroid=centroid,
            resolution=resolution,
            origin=origin,
        )
        stack_ellipsoids[instance] = ellipsoid

        surface_area = ellipsoid_surface_area(ellipsoid=ellipsoid)
        stack_surface_areas[instance] = surface_area

        volume = ellipsoid_volume(ellipsoid=ellipsoid)
        stack_volumes[instance] = volume

    properties = InstanceProperties(
        source_name=instance_stack.name,
        labels=inst_indices.labels,
        n_voxels=stack_n_voxels,
        equivalent_sphere_diameters=stack_eq_diameters.tolist(),
        centroids=Centroids(data=stack_centroids.tolist()),
        ellipsoids=BestFitEllipsoids(data=stack_ellipsoids.tolist()),
        surface_areas=EllipsoidSurfaceAreas(data=stack_surface_areas.tolist()),
        volumes=EllipsoidVolumes(data=stack_volumes.tolist()),
    )

    return properties


def process(yml_file: Path) -> bool:
    """
    Processes a directory of images specified in a YAML file, creates a semantic image stack,
    and performs instance analysis on all semantic label classes within the stack.
    Saves all instance stacks and properties into an HDF5 file containing the original semantic image stack.

    Parameters
    ----------
    yml_file : Path
        The fully pathed YAML file that specifies settings.

    Returns
    -------
    bool
        True if the process was successful, False otherwise.

    Examples
    --------
    >>> yml_file = Path("path/to/input.yml")
    >>> process(yml_file)
    True
    """

    semantic_stack = process_image_stack(yml_file)
    n_semantic_labels = np.unique(semantic_stack.data)

    # create h5
    h5_path = hio.image_to_voxel(yml_file)

    # start instance analysis
    db = ut.yaml_to_dict(yml_file)
    for instance_name in db["class_labels"]:

        # # check yml values match the labels in the semantic image stack
        # for semantic_label in n_semantic_labels:

        if not instance_analysis_included(settings=db, key=instance_name):
            continue

        print(f"Analyzing '{instance_name}' semantic label")

        # create the instance_stacks
        inst_stack = semantic_to_instance(
            semantic_stack=semantic_stack,
            instance_name=instance_name,
            instance_value=db["class_labels"][instance_name]["value"],
            min_feature_size=db["class_labels"][instance_name]["instance_analysis"][
                "min_feature_size"
            ],
        )

        # get instance_indices
        inst_indices = instance_indices(
            instance_stack=inst_stack,
        )
        if inst_stack.min_feature_size > 0:
            # overwrites instance_stack and inst_indices
            inst_stack, inst_indices = minimum_size_filter(
                initial_stack=inst_stack,
                initial_indices=inst_indices,
            )

        # add to h5
        hio.add_to_h5(
            inst_stack,
            h5_path=h5_path,
            h5_group="VoxelData",
        )
        print(f"\tlabelled '{instance_name}' voxel data added")

        # TODO not implemented instance indices yet
        # hio.add_to_h5(
        #     inst_indices,
        #     h5_path=h5_path,
        #     h5_group=f"{instance_name}",
        # )

        print(f"\tcalculating '{instance_name}' instance properties...")
        # calculate indepedendent feature properties
        inst_properties = instance_properties(
            instance_stack=inst_stack,
            inst_indices=inst_indices,
        )

        # add to h5
        hio.add_to_h5(
            inst_properties,
            h5_path=h5_path,
            h5_group=f"{instance_name}",
        )
        print(f"\t'{instance_name}' instance property data added")

        nearest_neighbors = nearest_neighbor_distance(
            centroids=inst_properties.centroids
        )
        hio.add_to_h5(
            nearest_neighbors,
            h5_path=h5_path,
            h5_group=f"{instance_name}",
        )
        print(f"\t'{instance_name}' neighborhood property data added")

    return True


def main():
    """
    Command-line interface for running instance analysis on image stacks.

    This function serves as the entry point for terminal-based access to the instance analysis module.
    It is invoked from the command line using the 'instance_analysis' command specified in pyproject.toml.
    The function processes a YAML input file to create a semantic image stack and performs instance analysis
    on all semantic label classes within the stack. The results, including instance stacks and properties,
    are saved into an HDF5 file.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Examples
    --------
    To run the instance analysis, use the following command in the terminal:
    $ instance_analysis path/to/input.yml
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", help="the .yml user input file")
    args = parser.parse_args()
    input_file = args.input_file
    process(input_file)

    print(f"\n{input_file} processed!")


if __name__ == "__main__":
    main()



"""The module test the voxel_to_mesh module.
"""

# python standard libraries
from pathlib import Path

# third-party libraries
import pytest

# local libraries
from recon3d import voxel_to_mesh as vm
from recon3d import static_test_paths as test_paths


def test_letter_f_3d():
    """Tests that the letter_f_3d.npy example can be run."""
    path_yml = test_paths.LETTER_F_3D_YML
    path_exo = test_paths.LETTER_F_3D_EXO

    result = vm.voxel_to_mesh(yml_input_file=path_yml)

    assert result  # success

    # clean up temporary Exodus II file that was created in this test.
    try:
        # delete the file
        assert path_exo.is_file()
        path_exo.unlink()
    except FileNotFoundError:
        print(f"Temporary file does not exist: {path_exo}")
    except PermissionError:
        print(f"Permission denied: Unable to delete tempoary file: {path_exo}")
    except Exception as e:
        print(f"An error occurred: {e}")



"""Tests that the types defined in the module are tested."""

import recon3d.types as tt


def test_origin():
    """Test the origin is at (0.0, 0.0, 0.0)."""
    length_null = tt.Length(value=0.0, unit=tt.Units.MICRON)
    origin = tt.Origin(x0=length_null, y0=length_null, z0=length_null)
    assert type(origin).__name__ == "Origin"
    assert origin.x0.value == 0.0
    assert origin.y0.value == 0.0
    assert origin.z0.value == 0.0


def test_centroid():
    """Tests that the Centroid data type is as expected."""
    centroid = tt.Centroid(
        cx=tt.Length(value=1.0, unit=tt.Units.MICRON),
        cy=tt.Length(value=2.0, unit=tt.Units.MICRON),
        cz=tt.Length(value=3.0, unit=tt.Units.MICRON),
    )
    assert type(centroid).__name__ == "Centroid"
    assert centroid.cx.value == 1.0
    assert centroid.cy.value == 2.0
    assert centroid.cz.value == 3.0


# Below is a history breadcrumb in case we need it for CI/CD.

"""A module for testing types.

Modules, classes, and methods can be
created for testing purposes.

"""

# import unittest
# from recon3d.core import Scan


# class TestTypes(unittest.TestCase):
#     """Class to test types."""

#     def main(self):
#         """Main function for module-level testing functionality."""

#         # TODO: Locate these implementations?  Comment out until found or implemented.
#         # self.test_int()  # where are these implementations?
#         # self.test_float()

#     def test_str(self):
#         """Function to test string representation."""

#         metadata = {"height": 2, "width": 3, "images": None, "xpixel_size": None}
#         self.assertEqual(type(Scan(metadata).__str__()), type("a string"))


# if __name__ == "__main__":
#     unittest.main()
